from cubicweb import (
    _,
    wfutils,
)


set_property('ui.site-title', 'test app')


useraccount_wf_definition = {
    'etypes': 'UserAccount',
    'default': True,
    'initial_state': _('created'),
    'states': [_('created'), _('activated'), _('deactivated')],
    'transitions': {
        _('activate'): {
            'fromstates': [u'created', u'deactivated'],
            'tostate': u'activated',
            'requiredgroups': u'managers'
        },
        _('deactivate'): {
            'fromstates': [u'activated'],
            'tostate': u'deactivated',
            'requiredgroups': u'managers'
        },
    },
}
wfutils.setup_workflow(cnx, u'User account worflow',
                       useraccount_wf_definition)



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema HTTP Link header utils"""

import re

import six
from webob.multidict import MultiDict


def serialize_links(link_dict):
    """Serialize link_dict into a Link Header
    compliant string as defined by RFC 5988
    """
    ldos = []
    for rel, links in link_dict.dict_of_lists().items():
        for link in links:
            link['rel'] = rel
            href = link.pop('href')
            params = u'; '.join(u'{}="{}"'.format(k, v)
                                for k, v in sorted(link.items()))
            ldos.append({"href": href, "params": params})
    ldos.sort(key=lambda ldo: (ldo["href"], ldo["params"]))
    header = u', '.join(u'<{href}>; {params}'.format(**ldo) for ldo in ldos)
    # Return a "native" string per WSGI specification (i.e. bytes on Python2
    # but unicode on Python 3):
    #   https://www.python.org/dev/peps/pep-3333/#a-note-on-string-types
    if six.PY2:
        header = header.encode('latin1')
    return header


def _split_unquotted(delimiter, string):
    r"""Split a string using using 'delimiter' as separator
    when it does not split a quoted string:

    >>> _split_unquotted(',', 'a,"b,c",d')
    ['a', '"b,c"', 'd']

    A quoted string can contain an escaped quote:

    >>> _split_unquotted(',', r'"a","\"b,\""')
    ['"a"', '"\\"b,\\""']
    """
    # Match any quoted string
    # even quoted strings which contain escaped quotes
    quoted_string = r'"(?:\\.|[^"\\])*"'
    # Match any string which does not contain an unquotted delimiter
    parts_regex = r'((?:[^"{delim}]+|{quoted_string})+)'.format(
        delim=delimiter,
        quoted_string=quoted_string)
    return [val for val in re.findall(parts_regex, string)]


def parse_links(link_header):
    r"""Parse a link header and generate a MultiDict

    This function generates a MultiDict of links
    by parsing a 'Link' header's value:

    >>> from pprint import pprint
    >>> link_header = (
    ...     '</A>;rel="a";type="application/json",'
    ...     '<B1>;rel="b";type="text/xml",'
    ...     '<B2>;rel="b";type="text/xml"'
    ... )
    >>> links = parse_links(link_header)
    >>> pprint(links.getone('a'))
    {'href': '/A', 'type': 'application/json'}
    >>> pprint(links.getall('b'))
    [{'href': 'B1', 'type': 'text/xml'}, {'href': 'B2', 'type': 'text/xml'}]

    This implementation allows spaces between links and between parameters:

    >>> links = parse_links('</A>; rel="a" ,</B> ;rel="b" ;  type="text/plain"')
    >>> pprint(links.dict_of_lists())
    {'a': [{'href': '/A'}], 'b': [{'href': '/B', 'type': 'text/plain'}]}

    It's possible to quote parameter values...

    >>> links = parse_links('</X>; rel="foo,bar"')
    >>> pprint(links.dict_of_lists())
    {'foo,bar': [{'href': '/X'}]}

    ...and to escape quotes in quoted values:

    >>> links = parse_links('</X>; title="The \"quotation\"" ; rel="related"')
    >>> pprint(links.dict_of_lists())
    {'related': [{'href': '/X', 'title': 'The "quotation"'}]}

    If a parameter is present and has no value specified, its default value
    will be True:

    >>> links = parse_links('</X>; use_modal; rel="next"')
    >>> pprint(links.dict_of_lists())
    {'next': [{'href': '/X', 'use_modal': True}]}

    Link values with several relations can also be parsed:

    >>> links = parse_links('</X>; rel="a b c"')
    >>> pprint(links.dict_of_lists())
    {'a': [{'href': '/X'}], 'b': [{'href': '/X'}], 'c': [{'href': '/X'}]}
    """

    links = MultiDict()
    for link_value in _split_unquotted(',', link_header):
        # link_value looks like '<THE_HREF>; parm="val1 val2"; parm=val3; parm4'
        href, params = re.search(r'^\s*<([^>]+)>\s*;(.*)$', link_value).groups()
        link = {'href': href}
        for param in _split_unquotted(';', params):
            param = param.strip()
            try:
                # expect the parameter to be a key/value pair
                key, value = _split_unquotted('=', param)
            except ValueError:
                # the parameter has no value specified
                key = param
                value = True
            else:
                value = value.strip()
                # remove quotes if the value is a quoted string
                if value[0] == '"':
                    value = value[1:-1]

            link[key] = value
        for rel in link.pop('rel').split(' '):
            links.add(rel, link)
    return links



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from unittest import TestCase

from mock import MagicMock

from cubicweb_jsonschema.predicates import NoWildcardAcceptPredicate


class NoWildcardAcceptPredicateTC(TestCase):

    def test_one_value(self):
        request = MagicMock()
        request.accept = ['text/html', '*/*']
        p = NoWildcardAcceptPredicate('text/html', None)
        self.assertTrue(p(None, request))
        p = NoWildcardAcceptPredicate('application/json', None)
        self.assertFalse(p(None, request))

    def test_multiple_values(self):
        request = MagicMock()
        request.accept = ['text/html', '*/*']
        p = NoWildcardAcceptPredicate(['text/html', 'application/json'], None)
        self.assertTrue(p(None, request))
        p = NoWildcardAcceptPredicate(
            ['application/json', 'application/octet-stream'], None)
        self.assertFalse(p(None, request))


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""Mappers with "jsonschema.collection" regid."""

from six import text_type

from logilab.common.registry import yes
from cubicweb import (
    neg_role,
    Unauthorized,
)
from cubicweb.predicates import (
    match_kwargs,
)

from cubicweb_jsonschema import (
    CREATION_ROLE,
)

from .base import (
    JSONSchemaMapper,
    JSONSchemaDeserializer,
    JSONSchemaSerializer,
    ProtectedDict,
    add_descriptive_metadata,
    add_links,
    object_schema,
)
from .predicates import (
    _for_workflowable_entity,
    yams_component_target,
    yams_final_rtype,
)
from . import _utils


__all__ = [
    'CollectionMapper',
    'EntityCollectionMapper',
    'RelatedCollectionMapper',
    'CollectionItemMapper',
    'CompositeRelationItemMapper',
]


@JSONSchemaSerializer.register
class CollectionMapper(JSONSchemaMapper):
    """Mapper for a collection of entities."""
    __regid__ = 'jsonschema.collection'

    def __repr__(self):
        return '<{0.__class__.__name__}>'.format(self)

    @add_links
    def schema_and_definitions(self, schema_role=None):
        if schema_role == CREATION_ROLE:
            return self._submission_schema_and_definitions()
        return self._array_schema(schema_role=schema_role)

    def _submission_schema_and_definitions(self):
        # Disallow submission as we have no knowledge of target entity type in
        # this mapper.
        return False, {}

    def _array_schema(self, schema_role=None):
        item_mapper = self.select_mapper(
            'jsonschema.item', **self.cw_extra_kwargs)
        items_schema, items_defs = item_mapper.schema_and_definitions(
            schema_role)
        schema = {
            'type': 'array',
            'items': items_schema,
        }
        return add_descriptive_metadata(schema, self), items_defs

    def links(self, schema_role=None):
        """Yield Link appobjects matching regid and selection context of this
        mapper if schema_role is None.
        """
        if schema_role is not None:
            return
        for link in super(CollectionMapper, self).links(
                schema_role=schema_role):
            yield link

    def serialize(self, entities):
        """Return a list of collection item representing each entity in
        `entities`.
        """
        def mapper(entity):
            """Select jsonschema.item mapper using entity's type."""
            return self.select_mapper(
                'jsonschema.item',
                etype=entity.cw_etype, **self.cw_extra_kwargs)

        return [mapper(entity).serialize(entity) for entity in entities]


@JSONSchemaDeserializer.register
class EntityCollectionMapper(CollectionMapper):
    """Mapper for a collection of entities of a given type."""
    __regid__ = 'jsonschema.collection'
    __select__ = CollectionMapper.__select__ & match_kwargs('etype')

    def __repr__(self):
        return '<{0.__class__.__name__} etype={0.etype}>'.format(self)

    @property
    def etype(self):
        return self.cw_extra_kwargs['etype']

    @property
    def title(self):
        """Title of the collection, plural form of entity type."""
        return self._cw._('{}_plural').format(self.etype)

    def _submission_schema_and_definitions(self):
        """Delegate generation of schema and definitions to the "entity"
        mapper corresponding to the entity type in this collection.
        """
        mapper = self.select_mapper(
            'jsonschema.entity', **self.cw_extra_kwargs)
        return mapper.schema_and_definitions(schema_role=CREATION_ROLE)

    def values(self, instance):
        mapper = self.select_mapper(
            'jsonschema.entity', **self.cw_extra_kwargs)
        return mapper.values(instance)

    def serialize(self, entities):
        """Return a list of collection item representing each entity in
        `entities`.
        """
        mapper = self.select_mapper('jsonschema.item', **self.cw_extra_kwargs)
        return [mapper.serialize(entity) for entity in entities]


class RelatedCollectionMapper(EntityCollectionMapper):
    """Mapper for a collection of entities through an *inlined* relation."""
    __select__ = (
        match_kwargs('entity', 'rtype', 'role')
        & yams_component_target()
    )

    def __repr__(self):
        return ('<{0.__class__.__name__}'
                ' rtype={0.rtype} role={0.role}>'.format(self))

    @property
    def role(self):
        return self.cw_extra_kwargs['role']

    @property
    def rtype(self):
        return self.cw_extra_kwargs['rtype']

    @property
    def title(self):
        """Title of the collection, name of the relation."""
        return self._cw._(self.rtype if self.role == 'subject'
                          else self.rtype + '-object')

    def _submission_schema_and_definitions(self):
        """Delegate generation of schema and definitions to the "entity"
        mapper selected with possible target of `rtype`, `role` bound to this
        mapper.
        """
        rschema = self._cw.vreg.schema[self.rtype]
        target_types = rschema.targets(role=self.role)
        assert len(target_types) == 1, \
            'cannot handle multiple target types in {}'.format(self)
        target_type = target_types[0]
        entity = self.cw_extra_kwargs['entity']
        mapper = self.select_mapper(
            'jsonschema.entity', etype=target_type,
            rtype=self.rtype, role=neg_role(self.role),
            target_types={entity.cw_etype},
        )
        return mapper.schema_and_definitions(schema_role=CREATION_ROLE)


class NonInlinedRelatedCollectionMapper(RelatedCollectionMapper):
    """Mapper for a collection of entities through a non-*inlined* relation."""
    __select__ = (
        match_kwargs('entity', 'rtype', 'role')
        & ~yams_component_target()
    )

    def _submission_schema_and_definitions(self):
        """Return schema and definitions accounting for constraints on
        possible targets of `rtype`, `role` relation information for `entity`
        bound to this mapper.
        """
        entity = self.cw_extra_kwargs['entity']
        rschema = self._cw.vreg.schema[self.rtype]
        # XXX similar loop in ETypeRelationItemMapper.relation_targets().
        ids = []
        for target_type in rschema.targets(role=self.role):
            try:
                rset = entity.unrelated(
                    self.rtype, target_type, role=self.role)
            except Unauthorized:
                continue
            for target in rset.entities():
                ids.append({
                    'type': 'string',
                    'enum': [text_type(target.eid)],
                    'title': target.dc_title(),
                })
        if not ids:
            return False, None
        properties = {
            'id': {
                'oneOf': ids,
            },
        }
        schema = object_schema(properties, required=['id'])
        return add_descriptive_metadata(schema, self), None


@JSONSchemaSerializer.register
class CollectionItemMapper(JSONSchemaMapper):
    """Mapper for an item of a collection."""
    __regid__ = 'jsonschema.item'
    __select__ = yes()

    @add_links
    def schema_and_definitions(self, schema_role=None):
        """Return either a string schema or an object with "type", "id and
        "title" properties.
        """
        if schema_role == CREATION_ROLE:
            raise ValueError(
                '{} is not appropriate for submission role'.format(self))
        schema = {
            'type': 'object',
            'properties': {
                'type': {
                    'type': 'string',
                },
                'id': {
                    'type': 'string',
                },
                'title': {
                    'type': 'string',
                },
            },
        }
        return add_descriptive_metadata(schema, self), {}

    def links(self, schema_role=None, **kwargs):
        kwargs['anchor'] = '#'
        return super(CollectionItemMapper, self).links(
            schema_role=schema_role, **kwargs)

    @staticmethod
    def serialize(entity):
        """Return a dictionary with entity represented as a collection item."""
        return {
            'type': entity.cw_etype.lower(),
            'id': text_type(entity.eid),
            'title': entity.dc_title(),
        }


class TrInfoCollectionMapper(EntityCollectionMapper):
    """Mapper for a collection of TrInfo associated with an entity."""
    __select__ = (
        match_kwargs({'etype': 'TrInfo'})
        & _for_workflowable_entity()
    )

    def values(self, *args):
        # This is handled by the view.
        raise NotImplementedError()

    def serialize(self):
        entity = self.cw_extra_kwargs['for_entity']
        wfobj = entity.cw_adapt_to('IWorkflowable')
        entities = wfobj.workflow_history
        return super(TrInfoCollectionMapper, self).serialize(entities)


class CompositeRelationItemMapper(CollectionItemMapper):
    """Mapper for items target of a composite relation."""
    __select__ = (
        ~yams_final_rtype()
        & yams_component_target()
        & (match_kwargs('etype') | match_kwargs('entity'))
        & match_kwargs('target_types')
    )

    # Per yams_component_target selector, 'rtype' and 'role' kwargs should be
    # present but 'target_types' is optional.
    def __init__(self, _cw, rtype, role, target_types=None,
                 etype=None, entity=None, **kwargs):
        super(CompositeRelationItemMapper, self).__init__(_cw, **kwargs)
        self.rtype = rtype
        self.role = role
        self.target_types = list(_utils.relation_target_types(
            _cw.vreg.schema, rtype, role, target_types))
        if etype is not None:
            self.etype = etype
        elif entity is not None:
            self.etype = entity.cw_etype

    def _inlined_relation(self):
        """Return True if relation from selection context is "inlined" per
        jsonschema uicfg.
        """
        rsection = self._cw.vreg['uicfg'].select('jsonschema', self._cw)
        tags = set([
            rsection.get(
                self.etype, self.rtype, target_type, self.role)
            for target_type in self.target_types
        ])
        if len(tags) != 1:
            raise ValueError(
                'inconsistent rtag for {} with {} as {} (targets: {})'.format(
                    self.rtype, self.etype, self.role, self.target_types)
            )
        # We actually check for both "inlined" and "hidden", since a
        # CompoundMapper may set a relation to "hidden" while we'll still want
        # it to be considered as "inlined" in this context. XXX
        return tags.pop() in ('inlined', 'hidden')

    def schema_and_definitions(self, schema_role=None):
        """Return schema and definitions for an item of a relation collection.

        In "creation" role, schema a oneOf with possible target types entity
        JSON schema. Otherwise, depending on wether the relation in "inlined"
        per jsonschema uicfg, either a full entity JSON Schema is returned or
        a genertic collection item JSON Schema.
        """
        if schema_role != CREATION_ROLE and not self._inlined_relation():
            return super(CompositeRelationItemMapper,
                         self).schema_and_definitions(schema_role=schema_role)
        items, definitions = [], ProtectedDict()
        for target_type in self.target_types:
            mapper = self.select_mapper(
                'jsonschema.entity', etype=target_type,
                rtype=self.rtype, role=neg_role(self.role),
                target_types={self.etype},
            )
            subschema, defs = mapper.schema_and_definitions(schema_role)
            items.append({
                '$ref': '#/definitions/{}'.format(target_type),
            })
            definitions[target_type] = subschema
            if defs:
                definitions.update(defs)
        nitems = len(items)
        if nitems == 0:
            return None, None
        elif nitems == 1:
            return items[0], definitions
        else:
            schema = {
                'oneOf': items,
            }
            return schema, definitions



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema Pyramid views for the JSON API of entities."""

import json
import traceback

from pyramid import httpexceptions
from pyramid.view import view_config
import six

from cubicweb import (
    ValidationError,
)
from cubicweb.web import (
    RequestError as _CWRequestError,
)
from .. import (
    CREATION_ROLE,
)
from ..resources import (
    ETypeResource,
    RootResource,
    WorkflowTransitionsResource,
)
from . import (
    describedby,
    entity_from_context,
    json_config,
    json_problem,
    LOG,
    pagination,
    up_link,
)


def _created(request, resource):
    location = request.resource_url(resource)
    request.response.location = location
    request.response.status_code = 201
    mapper = resource.mapper()
    return mapper.serialize()


def _json_body(request):
    """Return the JSON body found in `request` or raise "415 Unsupported Media
    Type".
    """
    exc_type = ValueError if six.PY2 else json.JSONDecodeError
    try:
        return request.json_body
    except exc_type:
        raise httpexceptions.HTTPUnsupportedMediaType()


@view_config(
    route_name='cubicweb-jsonschema.entities',
    context='cubicweb_jsonschema.resources.IRoot',
    request_method='GET',
    decorator=[describedby],
    permission='authenticated',
)
def get_root(context, request):
    """Site root, no data."""
    request.response.status_code = 204
    request.response.content_type = None
    request.response.allow = ['GET']
    return request.response


@json_config(
    route_name='cubicweb-jsonschema.entities',
    context='cubicweb_jsonschema.resources.IEntities',
    request_method='GET',
    decorator=[
        describedby,
        pagination,
        up_link,
    ],
)
def get_entities(context, request):
    """Render multiple entities in JSON format."""
    if context.has_perm('add'):
        request.response.allow = ['GET', 'POST']
    else:
        request.response.allow = ['GET']
    mapper = context.mapper()
    return mapper.serialize(context.rset.entities())


def allow_for_entity(entity):
    """Return a list of HTTP verbs that are Allow-ed for `entity`.
    """
    perm2verb = [
        ('read', 'GET'),
        ('update', 'PUT'),
        ('delete', 'DELETE'),
    ]
    return [v for p, v in perm2verb if entity.cw_has_perm(p)]


@json_config(
    route_name='cubicweb-jsonschema.entities',
    context='cubicweb_jsonschema.resources.IEntity',
    request_method='GET',
    decorator=[
        describedby,
        entity_from_context,
        up_link,
    ],
)
def get_entity(context, request):
    """Render a single entity in JSON format."""
    entity = context.entity
    request.response.allow = allow_for_entity(entity)
    mapper = context.mapper()
    return mapper.serialize()


@json_config(
    route_name='cubicweb-jsonschema.entities',
    context='cubicweb_jsonschema.resources.IEntities',
    request_method='POST',
)
def create_entity(context, request):
    """Create a new entity from JSON data."""
    # TODO In case of validation errors, it'd be better to give a JSON Schema
    # entry as a "pointer", would require selection context to be an
    # ETypeSchemaResource.
    mapper = context.mapper(CREATION_ROLE)
    instance = _json_body(request)
    values = mapper.values(instance)
    entity = request.cw_cnx.create_entity(mapper.etype, **values)
    request.cw_cnx.commit()
    LOG.info('created %s', entity)
    return _created(request, context[entity.eid])


@json_config(
    route_name='cubicweb-jsonschema.entities',
    context='cubicweb_jsonschema.resources.IEntity',
    request_method='PUT',
    decorator=[entity_from_context],
)
def update_entity(context, request):
    """Update an entity from JSON data."""
    entity = context.entity
    mapper = context.mapper()
    instance = _json_body(request)
    edited_props = list(instance)
    values = mapper.values(instance)
    entity.cw_set(**values)
    request.cw_cnx.commit()
    LOG.info('edited attributes "%s" of %s', ', '.join(edited_props), entity)
    request.response.location = request.resource_url(context)
    request.response.status_code = 200
    return mapper.serialize()


@view_config(
    route_name='delete_entity',
    context='cubicweb_jsonschema.resources.IEntity',
    request_method='DELETE',
    decorator=[entity_from_context],
)
def delete_entity(context, request):
    """Delete an entity."""
    entity = context.entity
    entity.cw_delete()
    request.cw_cnx.commit()
    LOG.info('deleted %s', entity)
    request.response.status_code = 204
    request.response.content_type = None
    return request.response


def allow_for_entity_relation(entity, rtype, role):
    eschema = entity.e_schema
    rdef = eschema.rdef(rtype, role)
    perm2verb = [
        ('read', 'GET'),
        ('add', 'POST'),
    ]
    if role == 'subject':
        kwargs = {'fromeid': entity.eid}
    else:
        kwargs = {'toeid': entity.eid}
    return [v for p, v in perm2verb
            if rdef.has_perm(entity._cw, p, **kwargs)]


@json_config(
    route_name='cubicweb-jsonschema.entities',
    context='cubicweb_jsonschema.resources.IRelatedEntities',
    request_method='GET',
    decorator=[
        describedby,
        pagination,
        up_link,
    ],
)
def get_related_entities(context, request):
    """Return a JSON document of entities target of a relationships."""
    entity = context.__parent__.rset.one()
    request.response.allow = allow_for_entity_relation(
        entity, context.rtype, context.role)
    mapper = context.mapper()
    return mapper.serialize(context.rset.entities())


@json_config(
    route_name='cubicweb-jsonschema.entities',
    context='cubicweb_jsonschema.resources.IRelatedEntities',
    request_method='POST',
)
def post_related_entities(context, request):
    """Insert relationships between entity bound to context and targets from
    request's JSON body.
    """
    entity = context.__parent__.rset.one()
    rtype, role = context.rtype, context.role
    vreg = request.registry['cubicweb.registry']
    mapper = vreg['mappers'].select(
        'jsonschema.relation', request.cw_request,
        entity=entity, rtype=rtype, role=role,
        resource=context,
    )
    payload = _json_body(request)
    if not isinstance(payload, dict):
        return json_problem(status=400, detail='expecting an object')
    instance = {rtype: [payload]}
    # Pass "None" as entity because we want to add a relation, not override
    # existing ones.
    values = mapper.values(None, instance)
    entity.cw_set(**values)
    request.cw_cnx.commit()
    target, = values[mapper.orm_rtype]
    # Depending to whether the relation is inlined or not we might get an
    # entity or and eid.
    eid = getattr(target, 'eid', target)
    try:
        related = context[eid]
    except KeyError:
        # Target may not be reachable through the "related" path as there's no
        # guarantee about this after commit() (e.g. hooks may update data).
        # We thus don't know what to return, hence a 204 No Content.
        request.response.status_code = 204
        request.response.content_type = None
        return request.response
    return _created(request, related)


@json_config(
    route_name='cubicweb-jsonschema.entities',
    context='cubicweb_jsonschema.resources.IRelatedEntity',
    request_method='GET',
    decorator=[
        describedby,
        up_link,
    ],
)
def get_related_entity(context, request):
    """Return a JSON document of an entity target of a relationship."""
    mapper = context.mapper()
    request.response.allow = allow_for_entity(context.entity)
    return mapper.serialize()


@json_config(
    route_name='cubicweb-jsonschema.entities',
    context='cubicweb_jsonschema.resources.IRelatedEntity',
    request_method='PUT',
)
def update_related_entity(context, request):
    """Update a related entity from JSON data."""
    mapper = context.mapper()
    instance = _json_body(request)
    edited_props = list(instance)
    values = mapper.values(instance)
    entity = context.entity
    entity.cw_set(**values)
    request.cw_cnx.commit()
    LOG.info('edited attributes "%s" of %s', ', '.join(edited_props), entity)
    request.response.status_code = 200
    request.response.location = request.resource_url(context)
    return mapper.serialize()


@view_config(
    route_name='delete_entity',
    context='cubicweb_jsonschema.resources.IRelatedEntity',
    request_method='DELETE',
)
def delete_relation(context, request):
    """Delete a relationship."""
    assert context.role == 'subject', 'can only delete subject relation'
    subj = context.__parent__.__parent__.rset.one()
    obj = context.entity
    request.cw_cnx.execute(
        'DELETE S {rtype} O WHERE S eid %(s)s, O eid %(o)s'.format(
            rtype=context.rtype),
        {'s': subj.eid, 'o': obj.eid})
    request.cw_cnx.commit()
    LOG.info('deleted <%s %s %s>', subj, context.rtype, obj)
    request.response.status_code = 204
    request.response.content_type = None
    return request.response


@json_config(
    route_name='cubicweb-jsonschema.entities',
    context=WorkflowTransitionsResource,
    request_method='GET',
    decorator=[
        describedby,
        up_link,
    ],
)
def get_entity_workflow_transitions(context, request):
    """Return a JSON document of workflow transitions of an entity."""
    mapper = context.mapper()
    return mapper.serialize()


@json_config(
    route_name='cubicweb-jsonschema.entities',
    context=WorkflowTransitionsResource,
    request_method='POST',
)
def post_entity_workflow_transition(context, request):
    """Pass a workflow transition for an entity."""
    payload = _json_body(request)
    trname = payload['name']
    comment = payload.get('comment')
    wfobj = context.for_entity.cw_adapt_to('IWorkflowable')
    wfobj.fire_transition(trname, comment=comment)
    request.cw_cnx.commit()
    request.response.status_code = 204
    request.response.content_type = None
    return request.response


@json_config(
    route_name='cubicweb-jsonschema.entities',
    context=ValidationError,
    # This view should be usable for unauthenticated users.
    permission=None,
)
def validation_failed(exc, request):
    """Exception view for ValidationError on JSON request."""
    LOG.info('%s encountered during processing of %s', exc, request)
    _ = request.cw_request._
    request.cw_cnx.rollback()
    exc.translate(_)
    invalid_params = []
    detail = []
    for rolename, value in exc.errors.items():
        if rolename:
            invalid_params.append({'name': rolename, 'reason': value})
        else:
            detail.append(value)
    kwargs = {}
    if detail:
        kwargs['detail'] = '\n'.join(detail)
    if invalid_params:
        kwargs['invalid-params'] = invalid_params
    return json_problem(status=422, **kwargs)


@json_config(
    route_name='cubicweb-jsonschema.entities',
    context=_CWRequestError,
    # This view should be usable for unauthenticated users.
    permission=None,
)
def cubicweb_requesterror(exc, request):
    """Handler for CubicWeb's RequestError instances.

    It will return a RFC 7807 formatted error response from exception found in
    context.

    This view breaks backward-compatibility with CubicWeb's behaviour without
    cubicweb-jsonschema as the JSON response returned by CubicWeb is not
    application/problem+json.
    """
    if request.cw_cnx is not None:
        request.cw_cnx.rollback()
    return json_problem(status=int(exc.status), detail=str(exc))


@json_config(
    route_name='cubicweb-jsonschema.entities',
    context=Exception,
    # This view should be usable for unauthenticated users.
    permission=None,
)
def generic_error(exc, request):
    """Generic exception handler for exception (usually during "edition"
    operations).

    It will return a RFC 7807 formatted error response from exception found in
    context.
    """
    # Because this handler is wired on the "Exception" context, any error will
    # be catched, including the standard pyramid return exceptions.
    # In this case, just return it
    if isinstance(exc, httpexceptions.HTTPException):
        return exc
    if request.cw_cnx is not None:
        request.cw_cnx.rollback()
    LOG.exception('exception occurred while processing %s', request)
    detail = None
    cwconfig = request.registry['cubicweb.config']
    if cwconfig.debugmode or cwconfig.mode == 'test':
        detail = ''.join(
            traceback.format_exception_only(type(exc), exc)
        ).strip()
    return json_problem(status=500, detail=detail)


def includeme(config):
    config.include('cubicweb.pyramid.predicates')
    config.include('..predicates')
    config.add_route('delete_entity', '/{etype}/*traverse',
                     factory=ETypeResource.from_match('etype'),
                     request_method='DELETE',
                     match_is_etype='etype')
    config.add_route(
        'cubicweb-jsonschema.entities',
        '*traverse',
        factory=RootResource,
        strict_accept='application/json',
    )
    config.scan(__name__)



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""Mappers with "jsonschema.entity" regid."""

from cubicweb import (
    _,
    ValidationError,
)
from cubicweb.predicates import (
    match_kwargs,
    relation_possible,
)

from cubicweb_jsonschema import (
    CREATION_ROLE,
    EDITION_ROLE,
    VIEW_ROLE,
)

from .base import (
    JSONSchemaMapper,
    JSONSchemaDeserializer,
    JSONSchemaSerializer,
    ProtectedDict,
    add_descriptive_metadata,
    add_links,
    object_schema,
)
from .predicates import (
    _for_workflowable_entity,
)


__all__ = [
    'ETypeMapper',
    'TargetETypeMapper',
    'EntityMapper',
    'TargetEntityMapper',
    'WorkflowableEntityMapper',
]


@JSONSchemaDeserializer.register
class ETypeMapper(JSONSchemaMapper):
    """JSON Schema mapper for entity types."""
    __regid__ = 'jsonschema.entity'
    __select__ = match_kwargs('etype')

    @property
    def etype(self):
        return self.cw_extra_kwargs['etype']

    @property
    def entity(self):
        """Entity bound to this mapper, None for ETypeMapper."""
        return None

    @property
    def title(self):
        return self._cw._(self.etype)

    @add_links
    def schema_and_definitions(self, schema_role=None):
        properties = {}
        required = []
        definitions = ProtectedDict()

        def insert_property(name, subschema, defs, mapper):
            if subschema is None:
                return
            properties[name] = subschema
            if defs:
                definitions.update(defs)
            if mapper.required(schema_role):
                required.append(name)

        for mapper in self._object_mappers():
            subschema, defs = mapper.schema_and_definitions(schema_role)
            insert_property(mapper.name, subschema, defs, mapper)

        for rtype, role, target_types in self.relations(schema_role):
            if target_types is None:
                target_types = self._rtype_target_types(rtype, role)
            assert isinstance(target_types, set), target_types
            mapper = self._relation_mapper(rtype, role, target_types)
            subschema, defs = mapper.schema_and_definitions(schema_role)
            insert_property(rtype, subschema, defs, mapper)

        schema = object_schema(properties, required)
        return add_descriptive_metadata(schema, self), definitions

    def links(self, schema_role=None):
        """Yield Link appobjects matching regid and selection context of this
        entity mapper along with selectable "jsonschema.relation" links for
        "subject" relations.
        """
        if schema_role is not None:
            return
        for linkcls in super(ETypeMapper, self).links():
            yield linkcls
        links_registry = self._cw.vreg['links']
        for rtype, role, __ in self.relations(VIEW_ROLE, section='related'):
            if role != 'subject':
                continue
            yield links_registry.select(
                'jsonschema.relation', self._cw,
                entity=self.entity, rtype=rtype, role=role,
            )

    def values(self, instance):
        """Return a dict with deserialized data from `instance` suitable for
        insertion in CubicWeb database.
        """
        entity = self.entity
        values = {}
        # Deserialize "jsonschema.object" mappers first.
        for mapper in self._object_mappers():
            values.update(mapper.values(entity, instance))
        # Then Yams relations.
        if entity is None:
            schema_role = CREATION_ROLE
        else:
            schema_role = EDITION_ROLE
        for rtype, role, target_types in self.relations(schema_role):
            mapper = self._relation_mapper(rtype, role, target_types)
            values.update(mapper.values(entity, instance))
        if instance:
            # All properties in "instance" should have been consumed at this
            # point.
            eid = entity.eid if entity is not None else None
            msg = _('unexpected properties: {}').format(', '.join(instance))
            raise ValidationError(eid, {None: msg})
        return values

    def relations(self, schema_role, section='inlined'):
        """Yield relation information tuple (rtype, role, targettypes)
        for given schema role in the context of bound entity type.

        Keyword argument `section` controls uicfg section to select relations
        in.
        """
        try:
            permission = {
                None: 'read',
                VIEW_ROLE: 'read',
                CREATION_ROLE: 'add',
                EDITION_ROLE: 'update',
            }[schema_role]
        except KeyError:
            raise ValueError('unhandled schema role "{0}" in {1}'.format(
                schema_role, self))
        entity = self.entity
        if entity is None:
            entity = self._cw.vreg['etypes'].etype_class(self.etype)(self._cw)
        rsection = self._cw.vreg['uicfg'].select(
            'jsonschema', self._cw, entity=entity)
        return rsection.relations_by_section(entity, section, permission)

    def _object_mappers(self):
        """Yield 'jsonschema.object' mapper instance selectable for entity
        bound to this mapper.
        """
        if 'jsonschema.object' not in self._cw.vreg['mappers']:
            return
        for mappercls in self._cw.vreg['mappers']['jsonschema.object']:
            if mappercls.__select__(mappercls, self._cw, etype=self.etype) > 0:
                yield mappercls(self._cw)

    def _rtype_target_types(self, rtype, role):
        rschema = self._cw.vreg.schema[rtype]
        return {t.type for t in rschema.targets(self.etype, role)}

    def _relation_mapper(self, rtype, role, target_types=None):
        return self.select_mapper(
            'jsonschema.relation',
            etype=self.etype, rtype=rtype, role=role, target_types=target_types)


class TargetETypeMapper(ETypeMapper):
    """Specialized version of :class:`ETypeMapper` selectable for an entity
    type as target of (`rtype`, `role`) relation.
    """
    __select__ = match_kwargs('etype', 'rtype', 'role')

    @property
    def rtype(self):
        return self.cw_extra_kwargs['rtype']

    @property
    def role(self):
        return self.cw_extra_kwargs['role']

    def relations(self, schema_role, section='inlined'):
        relations = super(TargetETypeMapper, self).relations(
            schema_role, section=section)
        for rtype, role, target_types in relations:
            if (rtype, role) == (self.rtype, self.role):
                continue
            yield rtype, role, target_types


@JSONSchemaSerializer.register
class EntityMapper(ETypeMapper):
    """JSON Schema mapper for an entity."""
    __select__ = match_kwargs('entity')

    @property
    def entity(self):
        """Live entity from selection context."""
        return self.cw_extra_kwargs['entity']

    @property
    def etype(self):
        return self.entity.cw_etype

    def _relation_mapper(self, rtype, role, target_types=None):
        return self.select_mapper(
            'jsonschema.relation',
            entity=self.entity, rtype=rtype,
            role=role, target_types=target_types,
            resource=None,
        )

    def serialize(self):
        """Return the serialized value entity bound to this mapper."""
        entity = self.entity
        entity.complete()
        data = {}
        for mapper in self._object_mappers():
            data[mapper.name] = mapper.serialize(entity)
        eschema = entity.e_schema
        for rtype, role, target_types in self.relations(VIEW_ROLE):
            relation_mapper = self._relation_mapper(rtype, role, target_types)
            if eschema.rdef(rtype, role=role, takefirst=True).final:
                value = relation_mapper.serialize(entity)
                if value is None:
                    continue
            else:
                rset = entity.related(
                    rtype, role, targettypes=tuple(target_types))
                if not rset:
                    continue
                value = relation_mapper.serialize(rset.entities())
            data[relation_mapper.orm_rtype] = value
        return data


class WorkflowableEntityMapper(EntityMapper):
    """Mapper for workflowable entity."""
    __select__ = (
        EntityMapper.__select__
        & relation_possible('in_state')
    )

    def schema_and_definitions(self, schema_role=None):
        schema, defns = super(
            WorkflowableEntityMapper, self).schema_and_definitions(
                schema_role=schema_role)
        properties = schema['properties']
        assert 'in_state' not in properties, schema
        properties['in_state'] = {
            'type': 'string',
            'title': self._cw._('state'),
            'readOnly': True,
        }
        return schema, defns

    def serialize(self):
        data = super(WorkflowableEntityMapper, self).serialize()
        wfentity = self.entity.cw_adapt_to('IWorkflowable')
        data['in_state'] = wfentity.state
        return data


class TargetEntityMapper(EntityMapper, TargetETypeMapper):
    """Specialized version of :class:`TargetETypeMapper` and
    :class:`EntityMapper` selectable for an *entity* as target of (`rtype`,
    `role`) relation.
    """
    __select__ = match_kwargs('entity', 'rtype', 'role')


class TrInfoEntityMapper(ETypeMapper):
    """Mapper for TrInfo entity types associated with an entity."""
    __select__ = (
        match_kwargs({'etype': 'TrInfo'})
        & _for_workflowable_entity()
    )

    @add_links
    def schema_and_definitions(self, schema_role=None):
        # For creation role, we return a custom schema (which is quite
        # unrelated to Yams schema) so that the end user sees a document with
        # the name of the transition (and a comment) instead of getting
        # exposed to the implementation details of CubicWeb workflows data
        # model.
        if schema_role != CREATION_ROLE:
            return super(TrInfoEntityMapper, self).schema_and_definitions(
                schema_role=schema_role)
        entity = self.cw_extra_kwargs['for_entity']
        wfobj = entity.cw_adapt_to('IWorkflowable')
        _ = self._cw._
        transitions_choice = [{'enum': [tr.name], 'title': _(tr.name)}
                              for tr in wfobj.possible_transitions()]
        if not transitions_choice:
            return False, None
        schema = {
            'type': 'object',
            'properties': {
                'name': {
                    'type': 'string',
                    'oneOf': transitions_choice,
                },
                'comment': {
                    'type': 'string'
                },
            },
            'required': ['name'],
        }
        return add_descriptive_metadata(schema, self), None



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""Predicates for Yams to JSON Schema mappers."""

from six import string_types

from logilab.common.registry import (
    Predicate,
    objectify_predicate,
)

from cubicweb.predicates import (
    ExpectedValuePredicate,
    PartialPredicateMixIn,
)


__all__ = [
    'yams_final_rtype',
    'yams_component_target',
    'yams_match',
    'partial_yams_match',
]


def _etype_from_context(kwargs):
    etype = kwargs.get('etype')
    if etype is None:
        entity = kwargs.get('entity')
        if entity is not None:
            etype = entity.cw_etype
    return etype


@objectify_predicate
def yams_final_rtype(cls, cnx, rtype=None, role=None, **kwargs):
    """Predicate that returns 1 when the supplied `rtype` is not final."""
    if rtype is None or role is None:
        return 0
    etype = _etype_from_context(kwargs)
    if etype is None:
        return 0
    rdef = cnx.vreg.schema[etype].rdef(rtype, role=role, takefirst=True)
    if rdef.final:
        return 1
    return 0


@objectify_predicate
def yams_component_target(cls, cnx, rtype=None, role=None, target_types=None,
                          **kwargs):
    """Predicate that returns 1 when the target entity types are component of
    the relation defined by `rtype` and `role` (i.e. the relation has
    composite set to `role`).
    """
    if rtype is None or role is None:
        return 0
    etype = _etype_from_context(kwargs)
    if etype is None:
        return 0
    component = None
    eschema = cnx.vreg.schema[etype]
    if target_types is None:
        rschema = cnx.vreg.schema[rtype]
        target_types = tuple(t.type for t in rschema.targets(etype, role=role))
    for target_type in target_types:
        rdef = eschema.rdef(rtype, role=role, targettype=target_type)
        _component = rdef.composite == role
        if component is None:
            component = _component
        elif not component == _component:
            cls.warning('component inconsistency accross target types')
            return 0
    return component


class yams_match(Predicate):
    """Predicate that returns a positive value when the supplied relation
    match parameters given as predicate argument.

    The more __call__ arguments match __init__'s ones, the higher the score
    is.
    """

    def __init__(self, etype=None, rtype=None, role=None, target_types=()):
        self.etype = etype
        self.rtype = rtype
        self.role = role
        if isinstance(target_types, string_types):
            target_types = {target_types}
        elif not isinstance(target_types, (set, frozenset)):
            target_types = frozenset(target_types)
        self.target_types = target_types

    def __call__(self, cls, cnx, rtype=None, role=None, **kwargs):
        etype = _etype_from_context(kwargs)
        score = 0
        for key in ('etype', 'rtype', 'role'):
            expected = getattr(self, key, None)
            if expected is not None:
                if locals()[key] != expected:
                    return 0
                score += 1
        target_types = kwargs.get('target_types', frozenset())
        assert isinstance(target_types, (set, frozenset)), \
            'yams_match must be called with a set as "target_types" argument'
        if self.target_types:
            if not target_types.issubset(self.target_types):
                return 0
            score += 1
        return score


class partial_yams_match(PartialPredicateMixIn, ExpectedValuePredicate):
    """Return non-zero if selected class's attributes `etype`, `rtype`, `role`
    and/or `target_types` match with keyword arguments from context.
    """

    match_attributes = ('etype', 'rtype', 'role', 'target_types')

    def __init__(self, *expected, **kwargs):
        super(partial_yams_match, self).__init__((), **kwargs)

    def complete(self, cls):
        self.expected = {}
        for name in self.match_attributes:
            if hasattr(cls, name):
                self.expected[name] = set([getattr(cls, name)])

    def _values_set(self, cls, req, **kwargs):
        values = {}
        for name in self.match_attributes:
            if name in kwargs:
                values[name] = kwargs[name]
        return values


@objectify_predicate
def _for_workflowable_entity(cls, cnx, for_entity=None, **kwargs):
    """Return 1 if `for_entity` context argument corresponds to a worflowable
    entity type.
    """
    if for_entity is not None:
        wfobj = for_entity.cw_adapt_to('IWorkflowable')
        if wfobj is not None:
            return 1
    return 0



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""Mappers with "jsonschema.relation" regid."""

import abc

import iso8601
from six import (
    add_metaclass,
    text_type,
)

from logilab.common.decorators import cachedproperty

from yams import BadSchemaDefinition, ValidationError
from yams.constraints import StaticVocabularyConstraint
from cubicweb import (
    Binary,
    Unauthorized,
    neg_role,
    _,
)
from cubicweb.predicates import (
    match_kwargs,
)

from cubicweb_jsonschema import (
    CREATION_ROLE,
    EDITION_ROLE,
    VIEW_ROLE,
    orm_rtype,
)

from .base import (
    JSONSchemaMapper,
    JSONSchemaDeserializer,
    JSONSchemaSerializer,
    add_descriptive_metadata,
    add_links,
    object_schema,
)
from .predicates import (
    _etype_from_context,
    yams_component_target,
    yams_final_rtype,
    yams_match,
)
from . import _utils


__all__ = [
    'BaseRelationMapper',
    'AttributeMapper',
    'StringMapper',
    'FloatMapper',
    'IntMapper',
    'BooleanMapper',
    'PasswordMapper',
    'DateMapper',
    'DatetimeMapper',
    'BytesMapper',
    'InlinedRelationMapper',
    'RelationMapper',
    'ETypeRelationItemMapper',
    'EntityRelationItemMapper',
]


class BaseRelationMapper(JSONSchemaMapper):
    """Base abstract class to fill the gap between a yams relation and it's json
    schema mapping.

    They should be selected depending on the relation (`etype`, `rtype`, `role`
    and optionaly `target_types`).
    """
    __regid__ = 'jsonschema.relation'
    __select__ = match_kwargs('rtype', 'role')
    __abstract__ = True

    @property
    def etype(self):
        """The entity type bound to this mapper."""
        return _etype_from_context(self.cw_extra_kwargs)

    def __init__(self, _cw, **kwargs):
        #: relation type name
        self.rtype = kwargs.pop('rtype')
        #: role of `etype` in relation
        self.role = kwargs.pop('role')
        #: possible target types of the relation (empty for attribute relations)
        self.target_types = list(_utils.relation_target_types(
            _cw.vreg.schema, self.rtype, self.role,
            kwargs.pop('target_types', None)))
        super(BaseRelationMapper, self).__init__(_cw, **kwargs)

    def __repr__(self):
        return ('<{0.__class__.__name__} etype={0.etype} rtype={0.rtype} '
                'role={0.role} target_types={0.target_types}>'.format(self))

    @cachedproperty
    def description(self):
        ntargets = len(self.target_types)
        if ntargets > 1:
            return None
        elif ntargets == 1:
            targettype = self.target_types[0]
        else:
            targettype = None
        eschema = self._cw.vreg.schema[self.etype]
        rdef = eschema.rdef(self.rtype, role=self.role,
                            targettype=targettype)
        if rdef.description:
            return self._cw._(rdef.description)

    @cachedproperty
    def title(self):
        if self.role == 'object':
            return self._cw._(self.rtype + '_object')
        return self._cw._(self.rtype)

    @cachedproperty
    def orm_rtype(self):
        return orm_rtype(self.rtype, self.role)

    def links(self, schema_role=None, **kwargs):
        """Yield Link appobjects matching regid and selection context of this
        mapper if schema_role is None.
        """
        if schema_role is not None:
            return
        for link in super(BaseRelationMapper, self).links(
                schema_role=schema_role, **kwargs):
            yield link


@JSONSchemaDeserializer.register
@JSONSchemaSerializer.register
@add_metaclass(abc.ABCMeta)
class AttributeMapper(BaseRelationMapper):
    """Abstract base abstract class to map attribute relation.

    Concrete class should implement the `json_type` property.
    """
    __abstract__ = True
    __select__ = yams_final_rtype() & match_kwargs('etype')

    @abc.abstractproperty
    def json_type(self):
        """JSON primitive type (e.g. "string", "number", etc.)"""

    #: JSON Schema "format" keyword for semantic validation.
    format = None

    @property
    def attr(self):
        """Relation definition for bound attribute."""
        return self._cw.vreg.schema[self.etype].rdef(self.rtype)

    def _constraint_mapper(self, cstr):
        mapper = self._cw.vreg['mappers'].select_or_none(
            'jsonschema.constraint', self._cw.vreg, self._cw._,
            self.etype, self.rtype, cstr)
        if mapper is not None:
            return mapper
        elif not isinstance(cstr, StaticVocabularyConstraint):
            self.warning('ignored %s on %s', cstr.type(), self.attr)
        return None

    def required(self, schema_role):
        """Return True if mapped property is *required*."""
        if schema_role in (CREATION_ROLE, EDITION_ROLE):
            return self.attr.cardinality[0] == '1'
        return False

    @add_links
    def schema_and_definitions(self, schema_role=None):
        schema = {
            'type': self.json_type,
        }
        if self.format is not None:
            schema['format'] = self.format
        if schema_role in (CREATION_ROLE, EDITION_ROLE):
            vocabulary_constraint = next(
                (cstr for cstr in self.attr.constraints
                 if isinstance(cstr, StaticVocabularyConstraint)), None)
            if vocabulary_constraint:
                # In presence of a vocabulary constraint, we wrap the field
                # into a oneOf field with a single-value 'enum', ignoring
                # other constraints.
                oneof_items = []
                for v in sorted(vocabulary_constraint.vocabulary()):
                    item_schema = schema.copy()
                    item_schema.update({
                        'enum': [v],
                        'title': self._cw._(v),
                    })
                    oneof_items.append(item_schema)
                schema = {
                    'oneOf': oneof_items,
                }
            else:
                for constraint in self.attr.constraints:
                    cstr_mapper = self._constraint_mapper(constraint)
                    if cstr_mapper is not None:
                        schema.update(cstr_mapper.json_schema(schema_role))
            if self.attr.default is not None:
                schema['default'] = self.attr.default

        return add_descriptive_metadata(schema, self), None

    def values(self, entity, instance):
        """Return a dictionary holding deserialized value for mapped attribute.

        If mapped attribute is absent from `instance` and `entity` is not
        None, the default value for attribute is returned as value of the
        dictionnary.
        """
        try:
            value = instance.pop(self.rtype)
        except KeyError:
            if entity is None:
                return {}
            rschema = entity.e_schema.rdef(self.rtype, self.role)
            value = rschema.default
        else:
            value = self._type(value)
            if entity is not None and getattr(entity, self.orm_rtype) == value:
                # Do not trigger update, if the value has not changed.
                # This is useful for PUT requests for which all fields should
                # be present even if they are not changed. By skipping
                # unchanged value, we avoid possible security check that would
                # be meaningless since the value is not changed.
                return {}
        return {self.orm_rtype: value}

    @staticmethod
    def _type(json_value):
        """Return properly typed value for use within a cubicweb's entity from
        given JSON value.

        Nothing to do by default.
        """
        return json_value

    def serialize(self, entity):
        value = getattr(entity, self.orm_rtype)
        if value is not None:
            return self._value(value)

    @staticmethod
    def _value(value):
        """Return the serializable value from attribute `value`."""
        return value


class StringMapper(AttributeMapper):
    """Attribute mapper for Yams' String type."""
    __select__ = yams_match(target_types='String')
    #:
    json_type = 'string'
    _type = text_type


class FloatMapper(AttributeMapper):
    """Attribute mapper for Yams' Float type."""
    __select__ = yams_match(target_types='Float')
    #:
    json_type = 'number'


class IntMapper(AttributeMapper):
    """Attribute mapper for Yams' Int and BigInt types."""
    __select__ = yams_match(target_types={'Int', 'BigInt'})
    #:
    json_type = 'integer'


class BooleanMapper(AttributeMapper):
    """Attribute mapper for Yams' Boolean type."""
    __select__ = yams_match(target_types='Boolean')
    #:
    json_type = 'boolean'


class PasswordMapper(AttributeMapper):
    """Attribute mapper for Yams' Password type."""
    __select__ = yams_match(target_types='Password')
    #:
    json_type = 'string'
    #:
    format = 'password'

    def required(self, schema_role):
        """Possibly return True unless in *edition* role."""
        if schema_role == EDITION_ROLE:
            return False
        return super(PasswordMapper, self).required(schema_role)

    def values(self, entity, instance):
        password_changed = self.orm_rtype in instance
        values = super(PasswordMapper, self).values(entity, instance)
        if entity is not None and not password_changed:
            # We don't want the Password value to be reset if it has not
            # changed.
            del values[self.orm_rtype]
        return values

    def serialize(self, entity):
        return None

    @staticmethod
    def _type(json_value):
        """Return an encoded string suitable for Password type."""
        return json_value.encode('utf-8')


class DateMapper(StringMapper):
    """Attribute mapper for Yams' Date type."""
    __select__ = yams_match(target_types=('Date'))
    #:
    format = 'date'

    @staticmethod
    def _type(value):
        """Return a datetime object parsed from ISO8601 `value` string."""
        return iso8601.parse_date(value)


class DatetimeMapper(DateMapper):
    """Attribute mapper for Yams' (TZ)Datetime type."""
    __select__ = yams_match(target_types=('Datetime', 'TZDatetime'))
    #:
    format = 'date-time'


class BytesMapper(StringMapper):
    """Attribute mapper for Yams' Bytes type."""
    __select__ = yams_match(target_types='Bytes')

    @staticmethod
    def _type(value):
        """Return a Binary containing `value`."""
        return Binary(value.encode('utf-8'))

    @staticmethod
    def _value(value):
        """Return a unicode string from Binary `value`."""
        return value.getvalue().decode('utf-8')


class _RelationMapper(BaseRelationMapper):
    """Abstract class for true relation (as opposed to attribute) mapper.
    """
    __abstract__ = True
    __select__ = ~yams_final_rtype()

    @add_links
    def schema_and_definitions(self, schema_role=None):
        item_mapper = self.select_mapper(
            'jsonschema.item',
            rtype=self.rtype, role=self.role,
            target_types=set(self.target_types),
            **self.cw_extra_kwargs)
        items_schema, defs = item_mapper.schema_and_definitions(schema_role)
        schema = {
            'type': 'array',
            'items': items_schema,
        }
        if schema_role in (CREATION_ROLE, EDITION_ROLE):
            cardinality = self._cardinality()
            if cardinality in '+1':
                schema['minItems'] = 1
            if cardinality in '?1':
                schema['maxItems'] = 1
        return add_descriptive_metadata(schema, self), defs

    def required(self, schema_role):
        if schema_role in (CREATION_ROLE, EDITION_ROLE):
            return self._cardinality() in '1+'
        return False

    def _cardinality(self):
        """Return role-cardinality if schema definition is consistent and
        raise BadSchemaDefinition otherwise.
        """
        rschema = self._cw.vreg.schema[self.rtype]
        cardinality = None
        for target_type in self.target_types:
            rdef = rschema.role_rdef(self.etype, target_type, self.role)
            card = rdef.role_cardinality(self.role)
            if cardinality is None:
                cardinality = card
            elif card != cardinality:
                raise BadSchemaDefinition(
                    'inconsistent {} cardinalities within {} relation '
                    'definitions'.format(self.role, self.rtype))
        return cardinality


@JSONSchemaDeserializer.register
@JSONSchemaSerializer.register
class InlinedRelationMapper(_RelationMapper):
    """Map relation as 'inlined', i.e. the target of the relation is
    created/edited along with its original entity.
    """
    __select__ = (_RelationMapper.__select__
                  & yams_component_target()
                  & (match_kwargs('etype') | match_kwargs('entity')))

    def values(self, entity, instance):
        # Would require knownledge of the target type from "instance",
        # but the generated JSON schema does not expose this yet.
        assert len(self.target_types) == 1, \
            'cannot handle multiple target types yet: {}'.format(
                self.target_types)
        target_type = self.target_types[0]
        try:
            values = instance.pop(self.rtype)
        except KeyError:
            if entity is None:
                return {}
            values = []
        if not isinstance(values, list):
            raise ValidationError(entity,
                                  {self.rtype: _('value should be an array')})
        if entity is not None:
            # if entity already exists, delete entities related through
            # this mapped relation
            for linked_entity in getattr(entity, self.orm_rtype):
                if linked_entity.cw_etype in self.target_types:
                    linked_entity.cw_delete()
        target_mapper = self.select_mapper(
            'jsonschema.entity', etype=target_type,
            rtype=self.rtype, role=neg_role(self.role),
            target_types={self.etype},
        )
        result = []
        for subinstance in values:
            subvalues = target_mapper.values(subinstance)
            result.append(self._cw.create_entity(target_type, **subvalues))
        return {self.orm_rtype: result}

    def serialize(self, related_entities):
        def serialize(entity):
            mapper = self.select_mapper(
                'jsonschema.entity', entity=entity,
                rtype=self.rtype, role=neg_role(self.role),
                target_types={self.etype},
            )
            return mapper.serialize()

        return [serialize(related) for related in related_entities]


@JSONSchemaDeserializer.register
@JSONSchemaSerializer.register
class RelationMapper(_RelationMapper):
    """Map relation as 'generic', i.e. the target of the relation may be
    selected in preexisting possible targets.
    """
    __select__ = (_RelationMapper.__select__
                  & ~yams_component_target()
                  & (match_kwargs('etype') | match_kwargs('entity')))

    def values(self, entity, instance):
        try:
            values = instance.pop(self.rtype)
        except KeyError:
            return {}
        if entity is not None:
            entity.cw_set(**{self.orm_rtype: None})
        try:
            values = [int(x['id']) for x in values]
        except (TypeError, ValueError):
            msg = _('value should be an array of string-encoded integers')
            raise ValidationError(entity, {self.rtype: msg})
        return {self.orm_rtype: values}

    def serialize(self, related_entities):
        item_mapper = self.select_mapper(
            'jsonschema.item',
            rtype=self.rtype, role=self.role, **self.cw_extra_kwargs)
        return [item_mapper.serialize(entity) for entity in related_entities]


class ETypeRelationItemMapper(BaseRelationMapper):
    """Map items of a 'generic' relation for an non-existant entity."""
    __regid__ = 'jsonschema.item'
    __select__ = (_RelationMapper.__select__
                  & ~yams_component_target()
                  & match_kwargs('etype', 'rtype', 'role'))

    def relation_targets(self, schema_role):
        entity = self._cw.vreg['etypes'].etype_class(self.etype)(self._cw)
        potential_targets = []
        for target_type in self.target_types:
            try:
                potential_targets.extend(entity.unrelated(
                    self.rtype, target_type, self.role).entities())
            except Unauthorized:
                continue
        return potential_targets

    @add_links
    def schema_and_definitions(self, schema_role=None):
        ids = [
            {
                'type': 'string',
                'enum': [text_type(target.eid)],
                'title': target.dc_title(),
            }
            for target in self.relation_targets(schema_role)
        ]
        if not ids:
            return False, None
        properties = {
            'id': {
                'oneOf': ids,
            },
        }
        required = []
        if schema_role in (CREATION_ROLE, EDITION_ROLE):
            required.append('id')
        return object_schema(properties, required), None

    # XXX copy of CollectionItemMapper's method
    def links(self, schema_role=None, **kwargs):
        kwargs['anchor'] = '#'
        return super(ETypeRelationItemMapper, self).links(
            schema_role=schema_role, **kwargs)


@JSONSchemaSerializer.register
class EntityRelationItemMapper(ETypeRelationItemMapper):
    """Map items of a 'generic' relation for an existing entity."""
    __select__ = (_RelationMapper.__select__
                  & ~yams_component_target()
                  & match_kwargs('entity', 'rtype', 'role'))

    @property
    def entity(self):
        return self.cw_extra_kwargs['entity']

    def relation_targets(self, schema_role):
        if schema_role == VIEW_ROLE:
            return self.entity.related(
                self.rtype, self.role,
                targettypes=tuple(self.target_types)).entities()
        if schema_role == CREATION_ROLE:
            assert len(self.target_types) == 1, \
                'cannot handle multiple target types in {} for {}'.format(
                    self, schema_role)
            targettype = self.target_types[0]
            return self.entity.unrelated(
                self.rtype, targettype, role=self.role,
            ).entities()
        return super(EntityRelationItemMapper, self).relation_targets(
            schema_role)

    def serialize(self, entity):
        return {'id': str(entity.eid)}



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema Pyramid views for JSON Schema endpoints."""

from functools import wraps

import json

from pyramid import httpexceptions
from pyramid.renderers import JSON
from pyramid.view import view_config

from cubicweb.web.views import uicfg

from .. import (
    CREATION_ROLE,
    EDITION_ROLE,
    JSONSCHEMA_MEDIA_TYPE,
    VIEW_ROLE,
    resources,
)
from . import (
    describes,
)


def schema_role(*allowed):
    """View decorator that parses the "role" query string parameter and checks
    that it's in `allowed` values.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(context, request):
            role = request.params.get('role')
            if role is not None and role not in allowed:
                raise httpexceptions.HTTPBadRequest(
                    'invalid role: {0}'.format(role))
            return func(context, request)
        return wrapper
    return decorator


def jsonschema_config(**settings):
    for name, value in [
        ('name', 'schema'),
        ('route_name', 'cubicweb-jsonschema.schema'),
        ('accept', JSONSCHEMA_MEDIA_TYPE),
        ('renderer', 'jsonschema'),
        ('request_method', 'GET'),
        ('permission', 'authenticated'),
    ]:
        settings.setdefault(name, value)
    return view_config(**settings)


def links_sortkey(link):
    """key function for sorting a list of link description object."""
    return link['href'], link['rel']


@jsonschema_config(
    context='cubicweb_jsonschema.resources.IRoot',
    decorator=[describes],
)
def application_schema(context, request):
    """Schema view for the application."""
    vreg = request.registry['cubicweb.registry']

    def links():
        for eschema in vreg.schema.entities():
            if uicfg.indexview_etype_section.get(eschema) != 'application':
                continue
            etype = eschema.type
            etype_resource = context[etype]
            title = etype + '_plural'
            link = vreg['links'].select(
                'jsonschema.collection', request.cw_request, title=title)
            yield link.description_object(etype_resource)

    schema = {
        'type': "null",
        'links': sorted(links(), key=links_sortkey),
    }
    cnx = request.cw_cnx
    if cnx is not None:
        schema['title'] = cnx.property_value('ui.site-title')
    return schema


@jsonschema_config(
    context='cubicweb_jsonschema.resources.IEntities',
    decorator=[
        describes,
        schema_role(VIEW_ROLE, CREATION_ROLE),
    ],
)
def etype_schema(context, request):
    """Schema view for an entity type.

    In presence of a "role" query parameter, the JSON Schema for specified
    role is returned. Otherwise, the complete JSON Hyper Schema is returned.
    """
    role = request.params.get('role')
    mapper = context.mapper(role)
    return mapper.json_schema(role)


@jsonschema_config(
    context='cubicweb_jsonschema.resources.IEntity',
    decorator=[
        describes,
        schema_role(VIEW_ROLE, EDITION_ROLE),
    ],
)
def entity_schema(context, request):
    """Schema view for a live entity.

    In presence of a "role" query parameter, the JSON Schema for specified
    role is returned. Otherwise, the complete JSON Hyper Schema is returned.
    """
    role = request.params.get('role')
    mapper = context.mapper(role)
    return mapper.json_schema(role)


@jsonschema_config(
    context='cubicweb_jsonschema.resources.IRelatedEntities',
    decorator=[
        describes,
        schema_role(VIEW_ROLE, CREATION_ROLE),
    ],
)
def related_entities_schema(context, request):
    """Schema view for the collection of targets of a relation.
    """
    role = request.params.get('role')
    mapper = context.mapper(role)
    return mapper.json_schema(role)


@jsonschema_config(
    context='cubicweb_jsonschema.resources.IRelatedEntity',
    decorator=[
        describes,
        schema_role(VIEW_ROLE, EDITION_ROLE),
    ],
)
def related_entity_schema(context, request):
    """Schema view for a target entity of a relation.

    In presence of a "role" query parameter, the JSON Schema for specified
    role is returned. Otherwise, the complete JSON Hyper Schema is returned.
    """
    role = request.params.get('role')
    mapper = context.mapper(role)
    return mapper.json_schema(role)


@jsonschema_config(
    context=resources.WorkflowTransitionsResource,
    decorator=[
        describes,
        schema_role(VIEW_ROLE, CREATION_ROLE),
    ],
)
def entity_workflow_transitions(context, request):
    """Schema view for workflow transitions of an entity."""
    role = request.params.get('role')
    mapper = context.mapper(role)
    return mapper.json_schema(role)


def includeme(config):
    config.include('..predicates')
    config.add_route(
        'cubicweb-jsonschema.schema',
        '*traverse',
        factory=resources.RootResource,
        strict_accept=JSONSCHEMA_MEDIA_TYPE,
    )

    def jsonschema_dumps(value, **kwargs):
        if isinstance(value, dict):
            value.setdefault(
                '$schema',
                'http://json-schema.org/draft-06/schema#',
            )
        elif not isinstance(value, bool):
            raise ValueError(
                "JSON Schema value must be of object or boolean type")
        return json.dumps(value, **kwargs)

    config.add_renderer(
        'jsonschema',
        JSON(serializer=jsonschema_dumps),
    )
    config.scan(__name__)



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema unit tests for Pyramid JSON API."""

import base64
from datetime import date
import json
from unittest import (
    TestCase,
    skip,
)

import jsonschema
from mock import patch
from pyramid import (
    renderers,
    testing,
)
import six
from webob.multidict import MultiDict

from cubicweb import Binary, ValidationError
from cubicweb.pyramid.test import PyramidCWTest

from cubicweb_jsonschema import VIEW_ROLE
from cubicweb_jsonschema.links import parse_links


class JSONSchemaRendererTC(TestCase):

    def setUp(self):
        self.config = testing.setUp()
        self.config.include('cubicweb_jsonschema.api.schema')

    def test_metaschema_for_object(self):
        self.assertEqual(
            renderers.render('jsonschema', {}),
            '{"$schema": "http://json-schema.org/draft-06/schema#"}',
        )
        self.assertEqual(
            json.loads(renderers.render('jsonschema', {'type': 'string'})),
            {'$schema': 'http://json-schema.org/draft-06/schema#',
             'type': 'string'},
        )

    def test_no_metaschema_for_bool(self):
        for value in (False, True):
            self.assertEqual(renderers.render('jsonschema', value),
                             str(value).lower())

    def test_invalid_value(self):
        with self.assertRaises(ValueError):
            renderers.render('jsonschema', 'a')


class BaseTC(PyramidCWTest):

    settings = {
        'cubicweb.bwcompat': False,
        'pyramid.debug_notfound': True,
        'cubicweb.auth.authtkt.session.secret': 'x',
        'cubicweb.auth.authtkt.persistent.secret': 'x',
        'cubicweb.session.secret': 'x',
    }

    def includeme(self, config):
        config.include('cubicweb_jsonschema.api')


class EntitiesTC(BaseTC):

    @skip('todo')
    def test_post_json_file_upload(self):
        data = {
            'login': u'bob',
            'upassword': u'bob',
            'picture': [{
                u'data': u'data:text/xml;name=test;base64,{}'.format(
                    base64.b64encode('hello')),
            }],
        }
        self.login()
        resp = self.webapp.post_json('/cwuser/', data, status=201,
                                     headers={'Accept': 'application/json'})
        with self.admin_access.cnx() as cnx:
            entity = cnx.find('CWUser', login=u'bob').one()
            self.assertTrue(entity.picture)
            photo = entity.picture[0]
            self.assertEqual(photo.read(), 'hello')
        self.assertEqual(
            resp.location, 'https://localhost:80/CWUser/%d' % entity.eid)

    @skip('todo')
    def test_post_json_file_upload_badrequest(self):
        self.login()
        for rtype, value in [
            ('unknown', [{'data': u'who cares?'}]),
            ('picture', [{'data': u'badprefix:blah blah'}]),
            ('picture', {'data': u'not in a list'}),
        ]:
            data = {rtype: value}
            with self.subTest(**data):
                data[u'login'] = u'bob'
                data[u'upassword'] = u'bob'
                # Rely on "status=400" for test assertion.
                self.webapp.post_json('/CWUser/', data, status=400,
                                      headers={'Accept': 'application/json'})

    def test_get_related(self):
        with self.admin_access.cnx() as cnx:
            book = cnx.create_entity('Book', title=u'title')
            author = cnx.create_entity('Author', name=u'bob',
                                       reverse_author=book)
            cnx.commit()
        self.login('anon', 'anon')
        url = '/book/%s/author' % book.eid
        res = self.webapp.get(url, headers={'accept': 'application/json'})
        related = res.json
        expected = [{
            'id': str(author.eid),
        }]
        with self.admin_access.cnx() as cnx:
            collection_mapper = cnx.vreg['mappers'].select(
                'jsonschema.collection', cnx, etype='Author')
            jschema = collection_mapper.json_schema(VIEW_ROLE)
        jsonschema.validate(related, jschema)
        self.assertEqual(related, expected)

    def test_post_bad_mtype(self):
        self.login()
        url = '/book/'
        self.webapp.post(url, 'this is not json', status=415,
                         headers={'Accept': 'application/json'})

    def test_post_related_not_an_object(self):
        with self.admin_access.cnx() as cnx:
            book = cnx.create_entity(
                'Book', title=u'title',
                author=cnx.create_entity('Author', name=u'a'))
            cnx.commit()
        self.login()
        url = '/book/%s/topics' % book.eid
        res = self.webapp.post_json(url, ['ahah'], status=400,
                                    headers={'accept': 'application/json'})
        expected = {
            'status': 400,
            'title': 'Bad Request',
            'detail': 'expecting an object',
        }
        self.assertEqual(res.json, expected)

    def test_post_related_bad_identifier(self):
        with self.admin_access.cnx() as cnx:
            book = cnx.create_entity(
                'Book', title=u'title',
                author=cnx.create_entity('Author', name=u'a'))
            cnx.create_entity('Topic', name=u'top')
            cnx.commit()
        self.login()
        url = '/book/%s/topics' % book.eid
        res = self.webapp.post_json(url, {'id': 'a'}, status=422,
                                    headers={'accept': 'application/json'})
        expected = {
            'status': 422,
            'title': 'Unprocessable Entity',
            'invalid-params': [{
                'name': 'topics',
                'reason': 'value should be an array of string-encoded integers',
            }],
        }
        self.assertEqual(res.json, expected)

    def test_post_related_reverse_role(self):
        with self.admin_access.cnx() as cnx:
            author = cnx.create_entity('Author', name=u'bob')
            cnx.commit()
        self.login()
        url = '/author/%s/reverse-author' % author.eid
        data = {'title': 'my first book'}
        response = self.webapp.post_json(
            url, data, status=201,
            headers={'accept': 'application/json'})
        expected_instance = {
            'title': 'my first book',
            'author': [{'id': str(author.eid)}],
        }
        self.assertEqual(response.json, expected_instance)
        with self.admin_access.cnx() as cnx:
            rset = cnx.find('Book', title=u'my first book')
            self.assertTrue(rset)
            book = rset.one()
        expected = 'https://localhost:80/author/{}/reverse-author/{}/'.format(
            author.eid, book.eid)
        self.assertEqual(response.location, expected)
        # Test GET on related entity (could be a dedicated test).
        response = self.webapp.get(response.location, status=200,
                                   headers={'Accept': 'application/json'})
        self.assertEqual(response.json, expected_instance)

    def test_get_related_reverse_role(self):
        with self.admin_access.cnx() as cnx:
            author = cnx.create_entity('Author', name=u'bob')
            books = [
                cnx.create_entity('Book', title=u'1', author=author),
                cnx.create_entity('Book', title=u'2', author=author),
            ]
            cnx.commit()
        self.login()
        url = '/author/%s/reverse-author' % author.eid
        response = self.webapp.get(url, status=200,
                                   headers={'Accept': 'application/json'})
        expected = [
            {'id': str(entity.eid),
             'title': entity.dc_title(),
             'type': entity.cw_etype.lower()}
            for entity in books
        ]
        self.assertCountEqual(response.json, expected)

    def test_post_related_entity_notfound(self):
        self.login()
        url = '/book/999/topics'
        self.webapp.post_json(url, {}, status=404,
                              headers={'accept': 'application/json'})

    def test_post_related_bad_target(self):
        with self.admin_access.cnx() as cnx:
            book = cnx.create_entity(
                'Book', title=u'title',
                author=cnx.create_entity('Author', name=u'a'))
            cnx.commit()
            cwuser_eid = cnx.find('CWUser', login=u'admin')[0][0]
        self.login()
        url = '/book/%s/topics' % book.eid
        # TODO: status should be 4xx
        self.webapp.post_json(url, {'id': str(cwuser_eid)}, status=500,
                              headers={'accept': 'application/json'})

    def test_put_with_incomplete_data(self):
        """A PUT request *replaces* entity attributes, so if fields are
        missing from JSON request body, respective attributes are reset.
        """
        with self.admin_access.cnx() as cnx:
            entity = cnx.create_entity('Photo', data=Binary(b'plop'),
                                       flash=True,
                                       exposure_time=1.23)
            cnx.create_entity('Thumbnail', data=Binary(b'plip'),
                              reverse_thumbnail=entity)
            cnx.commit()
        self.login()
        url = '/photo/{}/'.format(entity.eid)
        self.webapp.put_json(url, {'data': 'plip', 'media_type': u'jpeg'},
                             headers={'Accept': 'application/json'})
        with self.admin_access.cnx() as cnx:
            entity = cnx.entity_from_eid(entity.eid)
            self.assertEqual(entity.data.getvalue(), b'plip')
            self.assertEqual(entity.media_type, u'jpeg')
            # 'thumbnail' absent from request body, we should get ().
            self.assertEqual(entity.thumbnail, ())
            self.assertFalse(cnx.find('Thumbnail'))
            # 'flash' has a default value, we should get this back.
            self.assertEqual(entity.flash, False)
            # 'exposure_time' absent from request body, we should get None.
            self.assertIsNone(entity.exposure_time)

    def test_get_related_sort(self):
        """Sort by modification_date ascending and descending"""
        with self.admin_access.cnx() as cnx:
            author = cnx.create_entity(
                'Author', name=u'bob')
            book1 = cnx.create_entity(
                'Book', title=u'1',
                publication_date=date(1976, 3, 1),
                author=author,
            )
            book2 = cnx.create_entity(
                'Book', title=u'1',
                publication_date=date(1977, 3, 1),
                author=author,
            )
            cnx.commit()

        self.login('anon', 'anon')
        ascending = [str(book1.eid), str(book2.eid)]
        descending = ascending[::-1]
        for sort, expected in [
            ('publication_date', ascending),
            ('-publication_date', descending),
        ]:
            with self.subTest(sort=sort):
                url = ('/author/%s/publications?sort=%s'
                       % (author.eid, sort))
                res = self.webapp.get(
                    url, headers={'accept': 'application/json'})
                entities = res.json
                self.assertEqual(len(entities), 2)
                ids = [d['id'] for d in entities]
                self.assertEqual(ids, expected)

    def test_validationerror_additionalproperties(self):
        data = {
            u'name': u'bob',
            u'born': u'1986',
        }
        self.login()
        res = self.webapp.post_json('/author/', data, status=422,
                                    headers={'Accept': 'application/json'})
        self.assertEqual(res.content_type, 'application/problem+json')
        self.assertEqual(res.json_body, {
            u'status': 422,
            u'title': u'Unprocessable Entity',
            u'detail': u'unexpected properties: born',
        })

    def test_validationerror_nosource(self):
        """Test validation_failed view with no specific source entry."""
        with self.admin_access.cnx() as cnx:
            gid = cnx.find('CWGroup', name=u'users')[0][0]
        data = {
            'login': u'bob',
            'upassword': u'pass',
            'in_group': [{'id': str(gid)}],
        }
        with patch('cubicweb.req.RequestSessionBase.create_entity',
                   side_effect=ValidationError(None, {None: u'unmapped'})):
            self.login()
            res = self.webapp.post_json('/cwuser/', data, status=422,
                                        headers={'Accept': 'application/json'})
            self.assertEqual(res.content_type, 'application/problem+json')
            self.assertEqual(res.json_body, {
                u'status': 422,
                u'title': u'Unprocessable Entity',
                u'detail': u'unmapped',
            })

    def test_validationerror_integrityerror(self):
        with self.admin_access.cnx() as cnx:
            gid = cnx.find('CWGroup', name=u'users')[0][0]
            self.create_user(cnx, u'bob')
            cnx.commit()
        data = {
            'login': 'bob',
            'upassword': 'pass',
            'in_group': [{'id': str(gid)}],
        }
        self.login()
        res = self.webapp.post_json('/cwuser/', data, status=422,
                                    headers={'Accept': 'application/json'})
        self.assertEqual(res.content_type, 'application/problem+json')
        self.assertEqual(res.json_body, {
            u'status': 422,
            u'title': u'Unprocessable Entity',
            u'detail': u'some relations violate a unicity constraint',
            u'invalid-params': [
                {u'name': u'login',
                 u'reason': u'login is part of violated unicity constraint'}],
        })

    def test_exception(self):
        with self.admin_access.cnx() as cnx:
            gid = cnx.find('CWGroup', name=u'users')[0][0]
        data = {
            'login': 'bob',
            'upassword': 'pass',
            'in_group': [{'id': str(gid)}],
        }
        self.login()
        msg = "invalid literal for int() with base 10: 'foo'"
        with patch('cubicweb.req.RequestSessionBase.create_entity',
                   side_effect=ValueError(msg)):
            res = self.webapp.post_json('/cwuser/', data, status=500,
                                        headers={'Accept': 'application/json'})
            self.assertEqual(res.content_type, 'application/problem+json')
            self.assertEqual(res.json_body, {
                u'status': 500,
                u'title': u'Internal Server Error',
                u'detail': u"ValueError: %s" % msg,
            })

    def test_exception_no_detail(self):
        data = {'title': 'thing'}
        self.login()
        # "detail" is only present in test/debug mode.
        self.config.mode = 'dummy'
        with patch('cubicweb.req.RequestSessionBase.create_entity',
                   side_effect=NotImplementedError()):
            res = self.webapp.post_json('/book/', data, status=500,
                                        headers={'Accept': 'application/json'})
            self.assertEqual(res.content_type, 'application/problem+json')
            self.assertEqual(res.json_body, {
                u'status': 500,
                u'title': u'Internal Server Error',
            })

    def test_pagination(self):
        with self.admin_access.cnx() as cnx:
            eids = [cnx.create_entity('Author', name=name).eid
                    for name in [u'Bob', u'Alice', u'Eve', u'John']]
            cnx.commit()

        self.login()
        res = self.webapp.get('/author/?limit=1',
                              headers={'Accept': 'application/json'})

        self.assertEqual([int(x["id"]) for x in res.json_body], [eids[-1]])

        links = parse_links(res.headers['Link'])

        self.assertEqual(links.getone('next'), {
            'href': '/author/?limit=1&offset=1',
            'type': 'application/json',
        })
        self.assertNotIn('prev', links)

        res = self.webapp.get('/author/?limit=1&offset=1',
                              headers={'Accept': 'application/json'})
        links = parse_links(res.headers['Link'])
        self.assertEqual([int(x["id"]) for x in res.json_body], [eids[-2]])

        expected_links = MultiDict({
            'next': {
                'href': '/author/?limit=1&offset=2',
                'type': 'application/json',
            },
            'prev': {
                'href': '/author/?limit=1&offset=0',
                'type': 'application/json',
            },
        })
        self.assertDictContainsSubset(expected_links, links)

        # last page
        res = self.webapp.get('/author/?limit=1&offset=3',
                              headers={'Accept': 'application/json'})
        links = parse_links(res.headers['Link'])
        self.assertEqual([int(x["id"]) for x in res.json_body], [eids[0]])

        expected_links = MultiDict({
            'prev': {
                'href': '/author/?limit=1&offset=2',
                'type': 'application/json',
            },
        })
        self.assertDictContainsSubset(expected_links, links)
        self.assertNotIn('next', links)

    def test_pagination_badrequest(self):
        self.login()

        res = self.webapp.get('/author/?limit=0', status=400,
                              headers={'Accept': 'application/json'})
        errors = res.json_body
        expected = {
            'status': 400,
            'detail': 'invalid limit: 0',
            'title': 'Bad Request',
        }
        self.assertEqual(errors, expected)

        res = self.webapp.get('/author/?limit=foo', status=400,
                              headers={'Accept': 'application/json'})
        errors = res.json_body
        expected = {
            'status': 400,
            'detail': "invalid limit (foo)",
            'title': 'Bad Request',
        }
        self.assertEqual(errors, expected)


class BWCompatTC(BaseTC):
    """Tests with 'cubicweb.bwcompat = true'."""

    settings = BaseTC.settings.copy()
    settings['cubicweb.bwcompat'] = True

    def test_cubicweb_requesterror(self):
        self.login()
        # Raises RemoteCallFailed (which has status=500).
        response = self.webapp.post_json('/ajax?fname=format_date', status=500,
                                         headers={'Accept': 'application/json'})
        self.assertEqual(response.content_type, 'application/problem+json')
        if six.PY2:
            detail = 'TypeError: format_date() takes exactly 2 arguments (1 given)'  # noqa: E501
        else:
            detail = 'TypeError: format_date() missing 1 required positional argument: \'strdate\''  # noqa: E501
        self.assertEqual(response.json_body, {
            'status': 500,
            'title': 'Internal Server Error',
            'detail': detail,
        })


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema predicates"""

from pyramid import predicates


class NoWildcardAcceptPredicate(predicates.AcceptPredicate):
    """Overrides accept predicate to eliminate match on "*/*" value.

    See https://github.com/Pylons/pyramid/issues/1264.
    """

    def __call__(self, context, request):
        request_accepts = [h for h in request.accept if h != '*/*']
        for accept_type in self.values:
            if accept_type in request_accepts:
                return True
        return False


def includeme(config):
    config.add_route_predicate('strict_accept', NoWildcardAcceptPredicate)
    config.scan(__name__)



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema unit tests for HTTP Link header utils"""

import doctest
from unittest import TestCase
from webob.multidict import MultiDict

from cubicweb_jsonschema.links import (serialize_links,
                                       parse_links)


class LinkHeaderTest(TestCase):

    def test_serialize(self):
        links = MultiDict()
        links.add('describedby', {'href': 'LINK1'})
        links.add('describedby', {'href': 'LINK2', 'title': u'\xe9'})
        links['up'] = {'href': 'LINK3'}

        expected = ('<LINK1>; rel="describedby", '
                    '<LINK2>; rel="describedby"; title="\xe9", '
                    '<LINK3>; rel="up"')
        self.assertEqual(serialize_links(links), expected)

    def test_parse(self):
        link_header = ('<LINK1>; rel="describedby",'
                       '<LINK2> ; rel="describedby" ,'
                       '<LINK3>;    rel="up" ')

        links = MultiDict()
        links.add('describedby', {'href': 'LINK1'})
        links.add('describedby', {'href': 'LINK2'})
        links['up'] = {'href': 'LINK3'}

        self.assertEqual(parse_links(link_header), links)


def load_tests(loader, tests, ignore):
    tests.addTests(
        doctest.DocTestSuite('cubicweb_jsonschema.links')
    )
    return tests


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""Base definitions for Yams to JSON Schema mappers."""

import abc
from functools import wraps

import six

from cubicweb.appobject import AppObject

from .._compat import ABCMeta


__all__ = [
    'JSONSchemaMapper',
    'JSONSchemaDeserializer',
    'JSONSchemaSerializer',
]


def add_descriptive_metadata(schema, mapper):
    """Add "title" and "description" metadata to `schema` if found on
    `mapper`.
    """
    for keyword in ('title', 'description'):
        if keyword not in schema:
            value = getattr(mapper, keyword, None)
            if value is not None:
                schema[keyword] = value
    return schema


def add_links(schema_and_definitions):
    """Method decorator that collects Hyper-Schema links and possibly add them
    to JSON Schema produced by `schema_and_definitions` method.
    """
    @wraps(schema_and_definitions)
    def wrapper(self, schema_role=None):
        schema, defns = schema_and_definitions(self, schema_role=schema_role)
        if self.resource is not None and isinstance(schema, dict):
            links = [l.description_object(self.resource)
                     for l in self.links(schema_role=schema_role)]
            # Sort links to get predictable results.
            links.sort(key=lambda l: (l['href'], l.get('rel')))
            if links:
                schema['links'] = links
        return schema, defns

    return wrapper


class ProtectedDict(dict):
    """A dict which keys cannot be overridden."""

    def __setitem__(self, key, value):
        if key in self:
            raise ValueError("key '{}' is already defined".format(key))
        return super(ProtectedDict, self).__setitem__(key, value)


def object_schema(properties, required=None):
    schema = {
        'type': 'object',
        'properties': properties,
        'additionalProperties': False,
    }
    if required:
        schema['required'] = required
    return schema


@six.add_metaclass(abc.ABCMeta)
class JSONSchemaMapper(AppObject):
    """Abstract base class for mappers between Yams schema definitions and
    JSON Schema documents.
    """
    __registry__ = 'mappers'

    def __init__(self, _cw, resource=None, **extra):
        """Initialize a JSON Schema mapper from a CubicWeb `_cw` and a Pyramid
        `resource`.
        """
        self.resource = resource
        super(JSONSchemaMapper, self).__init__(_cw, **extra)

    def select_mapper(self, regid, **ctx):
        """Return a mapper with `regid` selected with `**ctx`.

        The `resource` bound to this mapper is passed in selection context,
        unless specified in `**ctx`.
        """
        ctx.setdefault('resource', self.resource)
        return self._cw.vreg['mappers'].select(regid, self._cw, **ctx)

    def json_schema(self, schema_role=None):
        """Return the JSON Schema for bound Yams context.

        If `schema_role` is not None a plain JSON Schema document is returned
        following specified *role* ("view", "creation" or "edition").
        Otherwise, a JSON Hyper Schema document with *links* is returned.
        """
        schema, definitions = self.schema_and_definitions(schema_role)
        if definitions:
            assert 'definitions' not in schema
            schema['definitions'] = definitions
        return schema

    @abc.abstractmethod
    @add_links
    def schema_and_definitions(self, schema_role):
        """Return the JSON (sub-)Schema for bound Yams context along with a
        dict of subschemas to be inserted in the main JSON Schema's
        "definitions" property.
        """

    def links(self, schema_role=None, **kwargs):
        """Yield Link appobjects matching regid and selection context of this
        mapper.

        Extra keyword arguments are used for Link appobjects selection (in
        addition to mapper's section context).
        """
        args = (self._cw, )
        kwargs.update(self.cw_extra_kwargs)
        links_registry = self._cw.vreg['links']
        for linkcls in links_registry.get(self.__regid__, ()):
            if linkcls.__select__(linkcls, *args, **kwargs) > 0:
                yield links_registry.selected(linkcls, args, kwargs)


@six.add_metaclass(ABCMeta)
class JSONSchemaDeserializer(object):
    """Abstract mixin class for mappers capable of deserializing a JSON
    `instance` valid under mapped JSON Schema and insert respective data into
    CubicWeb's database.
    """

    @abc.abstractmethod
    def values(self, instance):
        """Return a dict holding deserialized data from `instance` suitable for
        insertion in CubicWeb database for mapped schema element.

        Items from `instance` dictionnary are pop-ed upon processing.
        """

    @classmethod
    def __subclasshook__(cls, subclass):
        if cls is JSONSchemaDeserializer:
            if any('values' in parent.__dict__
                   for parent in subclass.__mro__):
                return True
        return NotImplemented


@six.add_metaclass(ABCMeta)
class JSONSchemaSerializer(object):
    """Abstract mixin class for mappers capable of serializing data from
    CubicWeb's database as a JSON `instance` valid under mapped JSON Schema.
    """

    @abc.abstractmethod
    def serialize(self, *args):
        """Return the serialized value for mapped schema element."""

    @classmethod
    def __subclasshook__(cls, subclass):
        if cls is JSONSchemaSerializer:
            if any('serialize' in parent.__dict__
                   for parent in subclass.__mro__):
                return True
        return NotImplemented



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""Mappers with "jsonschema.object" regid."""

import abc

from six import (
    add_metaclass,
    text_type,
)

from logilab.common.decorators import classproperty

from cubicweb_jsonschema.views import jsonschema_section

from .base import (
    JSONSchemaMapper,
    JSONSchemaDeserializer,
    JSONSchemaSerializer,
    add_descriptive_metadata,
    object_schema,
    ProtectedDict,
)
from .predicates import (
    partial_yams_match,
)

__all__ = [
    'CompoundMapper',
]


@JSONSchemaDeserializer.register
@JSONSchemaSerializer.register
@add_metaclass(abc.ABCMeta)
class CompoundMapper(JSONSchemaMapper):
    """Mapper for a "compound" field gathering Yams relations into a
    dedicated JSON "object" to be inserted in "definitions" key of the JSON
    Schema document.

    The compound "object" will appear in the main JSON Schema document's
    properties under the `name` class attribute (defaults to class name).
    """
    __regid__ = 'jsonschema.object'
    __abstract__ = True
    __select__ = partial_yams_match()

    @abc.abstractproperty
    def etype(self):
        """entity type holding this compound field"""

    @abc.abstractproperty
    def relations(self):
        """sequence of relations gathered in this compound field"""

    @abc.abstractproperty
    def title(self):
        """title of the field"""

    # sequence of 'relation/role' computed from 'relations'
    _relations = ()

    @classproperty
    def name(cls):
        """name of the property to be inserted in the main JSON Schema
        document (defaults to class name).
        """
        return text_type(cls.__name__)

    @classmethod
    def __registered__(cls, reg):
        if not cls.relations:
            raise ValueError(
                '{} is missing a "relations" class attribute'.format(cls))
        # Compute relation/role pairs from relations
        relations = []
        for relinfo in cls.relations:
            try:
                rtype, role = relinfo
            except ValueError:
                rtype = relinfo
                role = 'subject'
            relations.append((rtype, role))
        cls._relations = tuple(relations)
        # Check for name or relation duplication
        for obj in reg[cls.__regid__]:
            if obj == cls:
                continue
            if obj.etype == cls.etype:
                if obj.name == cls.name:
                    # Make sure 'name' is unique amongst objects with
                    # 'jsonschema.object' regid.
                    raise ValueError('a class with name "{}" is already '
                                     'registered for {} entity type'.format(
                                         cls.name, cls.etype))
                # Prevent duplicate mapping of the same etype/rtype.
                common_rtypes = set(obj._relations) & set(cls._relations)
                if common_rtypes:
                    raise ValueError(
                        'duplicate relation mapping for "{}": {}'.format(
                            cls.etype,
                            ', '.join('-'.join(rel) for rel in common_rtypes),
                        )
                    )
        # Hide relations mapped to this document from etype JSON Schema.
        for rtype, role in cls._relations:
            if role == 'object':
                jsonschema_section.tag_object_of(
                    ('*', rtype, cls.etype), 'hidden')
            else:
                jsonschema_section.tag_subject_of(
                    (cls.etype, rtype, '*'), 'hidden')
        return super(CompoundMapper, cls).__registered__(reg)

    def required(self, schema_role):
        """Return False by default."""
        return False

    def schema_and_definitions(self, schema_role=None):
        properties = {}
        required = []
        definitions = ProtectedDict()
        for rtype, role in self._relations:
            rschema = self._cw.vreg.schema[rtype]
            target_types = {
                t.type for t in rschema.targets(self.etype, role)}
            mapper = self.select_mapper(
                'jsonschema.relation',
                etype=self.etype, rtype=rtype, role=role,
                target_types=target_types,
            )
            subschema, defs = mapper.schema_and_definitions(schema_role)
            if subschema is None:
                continue
            properties[mapper.orm_rtype] = subschema
            if mapper.required(schema_role):
                required.append(mapper.orm_rtype)
            if defs:
                definitions.update(defs)
        schema = {
            '$ref': '#/definitions/{}'.format(self.title),
        }
        definitions[self.title] = add_descriptive_metadata(
            object_schema(properties, required),
            self,
        )
        return schema, definitions

    def values(self, entity, instance):
        assert entity is None or entity.cw_etype == self.etype, \
            'cannot get "values" for {} with {}'.format(entity, self)
        try:
            values = instance.pop(self.name)
        except KeyError:
            if entity is None:
                return {}
            # Continue with an empty "value" that would get filled by compound
            # relation mappers with default or None values.
            values = {}
        for rtype, role in self._relations:
            mapper = self._relation_mapper(rtype, role)
            values.update(
                mapper.values(entity, values))
        return values

    def serialize(self, entity):
        assert entity.cw_etype == self.etype, \
            'cannot serialize {} with {}'.format(entity, self)
        data = {}
        for rtype, role in self._relations:
            mapper = self._relation_mapper(rtype, role)
            value = mapper.serialize(entity)
            if value is not None:
                data[mapper.orm_rtype] = value
        return data

    def _relation_mapper(self, rtype, role):
        rschema = self._cw.vreg.schema[rtype]
        target_types = {t.type for t in rschema.targets(self.etype, role)}
        return self.select_mapper(
            'jsonschema.relation', etype=self.etype,
            rtype=rtype, role=role, target_types=target_types,
        )



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-jsonschema resources tests"""

import unittest

from mock import (
    NonCallableMock,
    patch,
)

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb_jsonschema import resources


class ResourcesTC(unittest.TestCase):

    def tearDown(self):
        for cls in (resources.RootResource, resources.ETypeResource):
            cls._child_resources.clear()

    def test_metaclass(self):
        self.assertIsNot(resources.RootResource._child_resources,
                         resources.ETypeResource._child_resources)

    def test_pluggable_resource(self):

        @resources.RootResource.child_resource
        class SubResource(resources.TraversalResource):
            segment = 'segment'
            match = staticmethod(lambda r: True)

        request = object()
        resource = resources.RootResource(request)
        child_resource = resource['segment']
        self.assertIsInstance(child_resource, SubResource)
        self.assertIs(child_resource.__parent__, resource)
        self.assertIs(child_resource.request, request)

        # we've to patch resources.ETypeResource.__init__ else it will attempt
        # to access request.registry which is not set here (it would ends up
        # with a key error anyway)
        with patch.object(resources.ETypeResource, '__init__',
                          side_effect=KeyError):
            self.assertRaises(KeyError,
                              resource.__getitem__, 'unknown-segment')

    def test_multiple_matches_error(self):

        @resources.RootResource.child_resource
        class SubResource1(resources.TraversalResource):
            segment = 'segment'
            match = staticmethod(lambda r: True)

        @resources.RootResource.child_resource
        class SubResource2(resources.TraversalResource):
            segment = 'segment'
            match = staticmethod(lambda r: True)

        resource = resources.RootResource(None)
        with self.assertRaises(ValueError) as cm:
            resource.matching_resource('segment')

        self.assertIn('SubResource1', str(cm.exception))
        self.assertIn('SubResource2', str(cm.exception))


class WorkflowTransitionsResourceTC(CubicWebTC):

    def fake_request(self, cnx):
        registry = {'cubicweb.registry': self.vreg}
        return NonCallableMock(cw_cnx=cnx, params={},
                               registry=registry)

    def test_plugged_for_workflowable_entity(self):
        with self.admin_access.cnx() as cnx:
            root = resources.RootResource(self.fake_request(cnx))
            entity_resource = root['cwuser']['admin']
            try:
                entity_resource['workflow-transitions']
            except KeyError as e:
                self.fail(e)

    def test_not_plugged_for_non_workflowable_entity(self):
        with self.admin_access.cnx() as cnx:
            root = resources.RootResource(self.fake_request(cnx))
            entity_resource = root['cwgroup']['users']
            with self.assertRaises(KeyError):
                entity_resource['workflow-transitions']


if __name__ == '__main__':
    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema Pyramid "entities" resources definitions."""

import abc
from collections import defaultdict
from functools import wraps

from six import text_type, add_metaclass

from pyramid.decorator import reify
from pyramid.httpexceptions import (
    HTTPForbidden,
    HTTPNotFound,
)
from pyramid.security import (
    ALL_PERMISSIONS,
    Allow,
    Authenticated,
    DENY_ALL,
)
from zope.interface import (
    Attribute,
    Interface,
    implementer,
)

from .api import json_problem

from rql import nodes, TypeResolverException


def parent(init_method):
    """Decorator for resource class's __init__ method to bind the __parent__
    attribute to instance.
    """
    @wraps(init_method)
    def wrapper(self, *args, **kwargs):
        parent = kwargs.pop('parent', None)
        init_method(self, *args, **kwargs)
        self.__parent__ = parent

    return wrapper


def find_relationship(schema, etype, rtype, role='subject'):
    """Return (`rtype`, `role`) if `etype` has a relationship `rtype` else
    raise a ValueError.
    """
    eschema = schema[etype]
    for relinfo in eschema.relation_definitions():
        rschema, __, current_role = relinfo
        if rschema.final or rschema.meta:
            continue
        if rschema.type == rtype and current_role == role:
            return relinfo
    raise ValueError('-'.join([rtype, role]))


def need_cnx(func):
    """Method decorator checking if the request is bound to an authenticated
    user (for which request.cw_cnx is not None).
    """
    @wraps(func)
    def wrapped_method(self):
        cnx = self.request.cw_cnx
        if cnx is None:
            raise HTTPForbidden()
        return func(self, cnx)
    return wrapped_method


class _PluggableResourceType(type):
    """Metaclass for pluggable resources, ensuring they have their own
    _child_resources dictionary.
    """
    def __init__(cls, *args, **kwargs):
        super(_PluggableResourceType, cls).__init__(*args, **kwargs)
        cls._child_resources = defaultdict(list)


@add_metaclass(_PluggableResourceType)
class PluggableResource(object):
    """Base class for pluggable resources, which provides an API to register
    custom child resources.

    .. automethod: child_resource
    """

    @classmethod
    def child_resource(cls, child_resource):
        """Register `child_resource`, that should be a class inheriting from
        :class:`TraversalResource`.

        It may be used as a class decorator.
        """
        cls._child_resources[child_resource.segment].append(child_resource)
        return child_resource

    def matching_resource(self, value):
        """Return the resource matching `value` child segment or None if no one
        matches.

        If there are more than on matching resource, raises `ValueError`.
        """
        matches = []
        for child_resource in self._child_resources[value]:
            if child_resource.match(self):
                matches.append(child_resource(self.request, self))

        if matches:
            if len(matches) > 1:
                msg = 'more than one matching resources: {}'.format(matches)
                raise ValueError(msg)
            return matches[0]
        return None

    def __getitem__(self, value):
        # Make sure parent class has no __getitem__ as we do not call it here.
        # If it happens, we'll need to call it here.
        assert not hasattr(super(PluggableResource, self), '__getitem__')
        child_resource = self.matching_resource(value)
        if child_resource is None:
            raise KeyError(value)
        return child_resource


@add_metaclass(abc.ABCMeta)
class TraversalResource(object):
    """Abstract base class for custom child resource. Concrete classes should
    specify both `segment` class attribute and `match` method.
    """
    #: string indicating the matched segment,
    segment = None

    @abc.abstractmethod
    def match(parent_resource):
        """Take the parent resource's instance as argument and returning `True`
        if this resource applies to this context or `False` otherwise.
        """

    def __init__(self, request, parent):
        self.request = request
        self.__parent__ = parent
        self.__name__ = self.segment


class IResource(Interface):
    """Base interface for resource classes."""
    __parent__ = Attribute("Parent resource object")
    __name__ = Attribute("Resource name")

    def __getitem__(value):
        raise KeyError(value)


class IMapped(IResource):
    """Interface for resources associated with a mapper."""

    def mapper(schema_role=None):
        """Return the mapper associated with this resource."""


class IRoot(IResource):
    """Root resource interface"""


@implementer(IRoot)
class RootResource(PluggableResource):
    """Root resources for entities."""
    __parent__ = None
    __name__ = ''
    __acl__ = [
        # We need Authenticated because we rely on request.cw_cnx to be set
        # (it's None for unauthenticated users). No distinction is made
        # between CubicWeb's anonymous user and other kind of users.
        (Allow, Authenticated, ALL_PERMISSIONS),
        DENY_ALL,
    ]

    def __init__(self, request):
        self.request = request

    def __getitem__(self, value):
        try:
            return super(RootResource, self).__getitem__(value)
        except KeyError:
            pass
        try:
            return ETypeResource(self.request, value, self)
        except KeyError:
            raise KeyError(value)


class Paginable(object):
    """A paginable resource mixin"""

    @property
    def _has_next(self):
        return getattr(self, '__has_next', False)

    @_has_next.setter
    def _has_next(self, value):
        setattr(self, '__has_next', value)

    @reify
    def limit(self):
        cnx = self.request.cw_cnx
        if cnx is not None:
            default_limit = cnx.property_value('navigation.page-size')
        else:
            default_limit = 40
        req_value = self.request.params.get('limit', default_limit)
        try:
            limit = int(req_value)
        except ValueError:
            detail = 'invalid limit ({})'.format(req_value)
            raise json_problem(status=400, detail=detail)
        if limit < 1:
            detail = 'invalid limit: {}'.format(limit)
            raise json_problem(status=400, detail=detail)
        return limit

    @reify
    def offset(self):
        req_value = self.request.params.get('offset', 0)
        try:
            return int(req_value)
        except ValueError:
            detail = 'invalid offset ({})'.format(req_value)
            raise json_problem(status=400, detail=detail)

    def prev_link(self):
        if self.offset > 0:
            previous_offset = max(self.offset - self.limit, 0)
            query = {
                "limit": self.limit,
                "offset": previous_offset,
            }
            href = self.request.resource_path(self, query=sorted(query.items()))
            return {"href": href, "type": "application/json"}
        return None

    def next_link(self):
        if self._has_next:
            next_offset = self.offset + self.limit
            query = {
                "limit": self.limit,
                "offset": next_offset,
            }
            href = self.request.resource_path(self, query=sorted(query.items()))
            return {"href": href, "type": "application/json"}
        return None

    def paginate(self, select, args=None):
        select.set_limit(self.limit+1)
        select.set_offset(self.offset)
        rql = select.as_string()
        rset = self.request.cw_cnx.execute(rql, args)
        if len(rset) == self.limit + 1:
            self._has_next = True
            rset = rset.limit(self.limit)
        return rset


class IEntities(IMapped):
    """Resource interface for a collection of entities."""

    def has_perm(action):
        """Return True if permission for `action` is granted."""
        return False


@implementer(IEntities)
class ETypeResource(Paginable, PluggableResource):
    """A resource class for an entity type."""

    @classmethod
    def from_match(cls, matchname):
        def factory(request):
            return cls(request, request.matchdict[matchname])
        return factory

    def __init__(self, request, etype, parent=None):
        vreg = request.registry['cubicweb.registry']
        self.request = request
        self.etype = vreg.case_insensitive_etypes[etype.lower()]
        self.cls = vreg['etypes'].etype_class(self.etype)
        if parent is None:
            parent = RootResource(request)
        self.__parent__ = parent
        self.__name__ = self.etype.lower()

    @property
    def title(self):
        return self.request.cw_request.__('{}_plural'.format(self.etype))

    def __getitem__(self, value):
        try:
            return super(ETypeResource, self).__getitem__(value)
        except KeyError:
            pass
        for attrname in ('eid', self.cls.cw_rest_attr_info()[0]):
            resource = EntityResource(self.request, self.cls,
                                      attrname, value, self)
            try:
                rset = resource.rset
            except HTTPNotFound:
                continue
            if rset.rowcount:
                return resource
        raise KeyError(value)

    @reify
    @need_cnx
    def rset(self, cnx):
        select = self.cls.fetch_rqlst(cnx.user)
        return self.paginate(select)

    def mapper(self, schema_role=None):
        vreg = self.request.registry['cubicweb.registry']
        return vreg['mappers'].select(
            'jsonschema.collection', self.request.cw_request,
            etype=self.etype, resource=self)

    def has_perm(self, action):
        vreg = self.request.registry['cubicweb.registry']
        return vreg.schema[self.etype].has_perm(self.request.cw_cnx, action)


class IEntity(IMapped):
    """Entity resource interface."""
    entity = Attribute("Entity bound to this resource")


@implementer(IEntity)
class EntityResource(PluggableResource):
    """A resource class for an entity. It provide method to retrieve an entity
    by eid.
    """

    def __init__(self, request, cls, attrname, name, parent):
        self.request = request
        self.cls = cls
        self.attrname = attrname
        self.__name__ = name
        self.__parent__ = parent

    @property
    def title(self):
        return self.rset.one().dc_title()

    def __getitem__(self, value):
        vreg = self.request.registry['cubicweb.registry']
        schema = vreg.schema
        try:
            return super(EntityResource, self).__getitem__(value)
        except KeyError:
            pass
        prefix = 'reverse-'
        if value.startswith(prefix):
            rtype, role = value[len(prefix):], 'object'
        else:
            rtype, role = value, 'subject'
        try:
            rschema, __, role = find_relationship(
                schema, self.cls.cw_etype, rtype, role=role)
        except ValueError:
            raise KeyError(value)
        else:
            return RelatedEntitiesResource(
                self.request, rschema.type, role, parent=self)

    @reify
    @need_cnx
    def rset(self, cnx):
        if self.cls is None:
            return cnx.execute('Any X WHERE X eid %(x)s',
                               {'x': int(self.__name__)})
        st = self.cls.fetch_rqlst(cnx.user, ordermethod=None)
        st.add_constant_restriction(st.get_variable('X'), self.attrname,
                                    'x', 'Substitute')
        if self.attrname == 'eid':
            try:
                rset = cnx.execute(st.as_string(), {'x': int(self.__name__)})
            except (ValueError, TypeResolverException):
                # conflicting eid/type
                raise HTTPNotFound()
        else:
            rset = cnx.execute(st.as_string(), {'x': text_type(self.__name__)})
        return rset

    def mapper(self, schema_role=None):
        vreg = self.request.registry['cubicweb.registry']
        entity = self.rset.one()
        return vreg['mappers'].select(
            'jsonschema.entity', self.request.cw_request,
            entity=entity, resource=self)


@implementer(IMapped)
@EntityResource.child_resource
class WorkflowTransitionsResource(TraversalResource):
    """Resource for workflow transitions of an entity."""
    segment = 'workflow-transitions'

    @classmethod
    def match(cls, parent_resource):
        entity = parent_resource.rset.one()
        return entity.cw_adapt_to('IWorkflowable') is not None

    @property
    def for_entity(self):
        return self.__parent__.rset.one()

    def mapper(self, schema_role=None):
        vreg = self.request.registry['cubicweb.registry']
        return vreg['mappers'].select(
            'jsonschema.collection', self.request.cw_request,
            etype='TrInfo', for_entity=self.for_entity, resource=self)


class IRelatedEntities(IMapped):
    """Related entities resource interface."""
    rtype = Attribute("Relation type name")
    role = Attribute("Role of parent resource in the relation")


@implementer(IRelatedEntities)
class RelatedEntitiesResource(Paginable, PluggableResource):
    """A resource wrapping entities related to the entity bound to
    `entity_resource` through `rtype`/`role`.
    """

    @parent
    def __init__(self, request, rtype, role, **kwargs):
        self.request = request
        self.rtype = rtype
        self.role = role
        if role == 'object':
            self.__name__ = 'reverse-' + rtype
        else:
            self.__name__ = rtype

    @property
    def title(self):
        return self.request.cw_request.__(
            self.rtype + ('_object' if self.role == 'object' else ''))

    @reify
    @need_cnx
    def rset(self, cnx):
        # May raise HTTPNotFound, this is probably fine (isn't it?)
        entity = self.__parent__.rset.one()
        vreg = self.request.registry['cubicweb.registry']
        # XXX Until https://www.cubicweb.org/ticket/12306543 gets done.
        rql = entity.cw_related_rql(self.rtype, role=self.role)
        args = {'x': entity.eid}
        select = vreg.parse(entity._cw, rql, args).children[0]
        mainvar = select.get_variable('X')
        sortterms = self.request.params.get('sort')
        if sortterms:
            self._handle_sort_terms(select, mainvar, sortterms)
        return self.paginate(select, args)

    @staticmethod
    def _handle_sort_terms(select, mainvar, sortterms):
        select.remove_sort_terms()
        for term in sortterms.split(','):
            if term.startswith('-'):
                asc = False
                term = term[1:]
            else:
                asc = True
            mdvar = select.make_variable()
            rel = nodes.make_relation(mainvar, term,
                                      (mdvar,), nodes.VariableRef)
            select.add_restriction(rel)
            select.add_sort_var(mdvar, asc=asc)

    def __getitem__(self, value):
        try:
            return super(RelatedEntitiesResource, self).__getitem__(value)
        except KeyError:
            pass
        entity = self.__parent__.rset.one()
        if self.role == 'object':
            mvar = 'S'
            args = {'s': value, 'o': entity.eid}
        else:
            mvar = 'O'
            args = {'s': entity.eid, 'o': value}
        rset = self.request.cw_cnx.execute(
            'Any {mvar} WHERE O eid %(o)s, S {rtype} O, S eid %(s)s'.format(
                mvar=mvar, rtype=self.rtype),
            args)
        if not rset:
            raise KeyError(value)
        return RelatedEntityResource(
            self.request, rset.one(), rtype=self.rtype, role=self.role,
            parent=self)

    def mapper(self, schema_role=None):
        vreg = self.request.registry['cubicweb.registry']
        entity = self.__parent__.rset.one()
        return vreg['mappers'].select(
            'jsonschema.collection', self.request.cw_request,
            entity=entity, rtype=self.rtype, role=self.role,
            resource=self)


class IRelatedEntity(IRelatedEntities):
    """Related entity resource."""
    entity = Attribute("Entity target of the relation")


@implementer(IRelatedEntity)
class RelatedEntityResource(PluggableResource):
    """A resource an entity related to another one."""

    def __init__(self, request, entity, rtype, role, parent):
        self.request = request
        self.entity = entity
        self.rtype = rtype
        self.role = role
        self.__parent__ = parent
        self.__name__ = text_type(entity.eid)

    @property
    def title(self):
        return self.entity.dc_title()

    def mapper(self, schema_role=None):
        vreg = self.request.registry['cubicweb.registry']
        parent = self.__parent__
        return vreg['mappers'].select(
            'jsonschema.entity', self.request.cw_request,
            entity=self.entity, rtype=parent.rtype, role=parent.role,
            resource=self)



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-jsonschema permissions tests"""

from cubicweb.pyramid.test import PyramidCWTest


class PermissionsTests(PyramidCWTest):

    settings = {
        'cubicweb.bwcompat': False,
        'pyramid.debug_notfound': True,
        'cubicweb.auth.authtkt.session.secret': 'x',
        'cubicweb.auth.authtkt.persistent.secret': 'x',
        'cubicweb.session.secret': 'x',
    }

    def includeme(self, config):
        config.include('cubicweb_jsonschema.api')
        # Disable anonymous-user from Pyramid side (the user still exists, but
        # we make it hidden from Pyramid) so that, without an explicit login,
        # the request is unauthenticated (instead of authenticated with an
        # "anon" user).
        del config.registry['cubicweb.anonymous_eid']

    def test_anon(self):
        self.login('anon', 'anon')
        self.webapp.get('/cwuser', status=200,
                        headers={'Accept': 'application/json'})

    def test_user(self):
        self.login()
        self.webapp.get('/cwuser', status=200,
                        headers={'Accept': 'application/json'})

    def test_nobody(self):
        for path in [
                '/',
                '/schema',
                '/cwuser',
                '/cwuser/987',  # does not exist, but still 403
                '/cwuser/schema',
        ]:
            if path.endswith('/schema'):
                mtype = 'application/schema+json'
            else:
                mtype = 'application/json'
            with self.subTest(path=path):
                self.webapp.get(path, status=403,
                                headers={'Accept': mtype})


if __name__ == '__main__':
    import unittest
    unittest.main()
